# -*- coding: utf-8 -*-
'''
Integration tests for the vault modules
'''

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals
import inspect
import time

# Import Salt Testing Libs
from tests.support.unit import skipIf
from tests.support.case import ModuleCase, ShellCase
from tests.support.helpers import destructiveTest, flaky
from tests.support.paths import FILES

# Import Salt Libs
import salt.utils.path


@destructiveTest
@skipIf(not salt.utils.path.which('dockerd'), 'Docker not installed')
@skipIf(not salt.utils.path.which('vault'), 'Vault not installed')
class VaultTestCase(ModuleCase, ShellCase):
    '''
    Test vault module
    '''
    count = 0

    def setUp(self):
        '''
        SetUp vault container
        '''
        if VaultTestCase.count == 0:
            config = '{"backend": {"file": {"path": "/vault/file"}}, "default_lease_ttl": "168h", "max_lease_ttl": "720h"}'
            self.run_state('docker_image.present', name='vault', tag='0.9.6')
            self.run_state(
                'docker_container.running',
                name='vault',
                image='vault:0.9.6',
                port_bindings='8200:8200',
                environment={
                    'VAULT_DEV_ROOT_TOKEN_ID': 'testsecret',
                    'VAULT_LOCAL_CONFIG': config,
                },
                cap_add='IPC_LOCK',
            )
            time.sleep(5)
            ret = self.run_function(
                'cmd.retcode',
                cmd='/usr/local/bin/vault login token=testsecret',
                env={'VAULT_ADDR': 'http://127.0.0.1:8200'},
            )
            login_attempts = 1
            # If the login failed, container might have stopped
            # attempt again, maximum of three times before
            # skipping.
            while ret != 0:
                self.run_state(
                    'docker_container.running',
                    name='vault',
                    image='vault:0.9.6',
                    port_bindings='8200:8200',
                    environment={
                        'VAULT_DEV_ROOT_TOKEN_ID': 'testsecret',
                        'VAULT_LOCAL_CONFIG': config,
                    },
                    cap_add='IPC_LOCK',
                )
                time.sleep(5)
                ret = self.run_function(
                    'cmd.retcode',
                    cmd='/usr/local/bin/vault login token=testsecret',
                    env={'VAULT_ADDR': 'http://127.0.0.1:8200'},
                )
                login_attempts += 1
                if login_attempts >= 3:
                    self.skipTest('unable to login to vault')
            ret = self.run_function(
                'cmd.retcode',
                cmd='/usr/local/bin/vault policy write testpolicy {0}/vault.hcl'.format(FILES),
                env={'VAULT_ADDR': 'http://127.0.0.1:8200'},
            )
            if ret != 0:
                self.skipTest('unable to assign policy to vault')
        VaultTestCase.count += 1

    def tearDown(self):
        '''
        TearDown vault container
        '''
        def count_tests(funcobj):
            return inspect.ismethod(funcobj) or \
                inspect.isfunction(funcobj) and \
                funcobj.__name__.startswith('test_')
        numtests = len(inspect.getmembers(VaultTestCase, predicate=count_tests))
        if VaultTestCase.count >= numtests:
            self.run_state('docker_container.stopped', name='vault')
            self.run_state('docker_container.absent', name='vault')
            self.run_state('docker_image.absent', name='vault', force=True)

    @flaky
    def test_sdb(self):
        assert self.run_function('sdb.set', uri='sdb://sdbvault/secret/test/test_sdb/foo', value='bar') is True
        assert self.run_function('sdb.get', arg=['sdb://sdbvault/secret/test/test_sdb/foo']) == 'bar'

    @flaky
    def test_sdb_runner(self):
        assert self.run_run('sdb.set sdb://sdbvault/secret/test/test_sdb_runner/foo bar') == ['True']
        assert self.run_run('sdb.get sdb://sdbvault/secret/test/test_sdb_runner/foo') == ['bar']

    @flaky
    def test_config(self):
        assert self.run_function('sdb.set', uri='sdb://sdbvault/secret/test/test_pillar_sdb/foo', value='bar') is True
        assert self.run_function('config.get', arg=['test_vault_pillar_sdb']) == 'bar'



# coding: utf-8

# Python libs
from __future__ import absolute_import

# Salt testing libs
from tests.support.unit import skipIf, TestCase
from tests.support.mock import NO_MOCK, NO_MOCK_REASON, patch, MagicMock
from tests.support.mixins import LoaderModuleMockMixin

# Salt libs
import salt.beacons.smartos_vmadm as vmadm

# Mock Results
MOCK_CLEAN_STATE = {'first_run': True, 'vms': []}
MOCK_VM_NONE = {}
MOCK_VM_ONE = {
    '00000000-0000-0000-0000-000000000001': {
        'state': 'running',
        'alias': 'vm1',
        'hostname': 'vm1',
        'dns_domain': 'example.org',
    },
}
MOCK_VM_TWO_STOPPED = {
    '00000000-0000-0000-0000-000000000001': {
        'state': 'running',
        'alias': 'vm1',
        'hostname': 'vm1',
        'dns_domain': 'example.org',
    },
    '00000000-0000-0000-0000-000000000002': {
        'state': 'stopped',
        'alias': 'vm2',
        'hostname': 'vm2',
        'dns_domain': 'example.org',
    },
}
MOCK_VM_TWO_STARTED = {
    '00000000-0000-0000-0000-000000000001': {
        'state': 'running',
        'alias': 'vm1',
        'hostname': 'vm1',
        'dns_domain': 'example.org',
    },
    '00000000-0000-0000-0000-000000000002': {
        'state': 'running',
        'alias': 'vm2',
        'hostname': 'vm2',
        'dns_domain': 'example.org',
    },
}


@skipIf(NO_MOCK, NO_MOCK_REASON)
class SmartOSImgAdmBeaconTestCase(TestCase, LoaderModuleMockMixin):
    '''
    Test case for salt.beacons.vmadm
    '''

    def setup_loader_modules(self):
        return {
            vmadm: {
                '__context__': {},
                '__salt__': {},
            }
        }

    def test_non_list_config(self):
        '''
        We only have minimal validation so we test that here
        '''
        config = {}

        ret = vmadm.validate(config)

        self.assertEqual(ret, (False, 'Configuration for vmadm beacon must be a list!'))

    def test_created_startup(self):
        '''
        Test with one vm and startup_create_event
        '''
        # NOTE: this should yield 1 created event + one state event
        with patch.dict(vmadm.VMADM_STATE, MOCK_CLEAN_STATE), \
             patch.dict(vmadm.__salt__, {'vmadm.list': MagicMock(return_value=MOCK_VM_ONE)}):

            config = [{'startup_create_event': True}]

            ret = vmadm.validate(config)
            self.assertEqual(ret, (True, 'Valid beacon configuration'))

            ret = vmadm.beacon(config)
            res = [{'alias': 'vm1',
                    'tag': 'created/00000000-0000-0000-0000-000000000001',
                    'hostname': 'vm1',
                    'dns_domain': 'example.org'},
                   {'alias': 'vm1',
                    'tag': 'running/00000000-0000-0000-0000-000000000001',
                    'hostname': 'vm1',
                    'dns_domain': 'example.org'}]
            self.assertEqual(ret, res)

    def test_created_nostartup(self):
        '''
        Test with one image and startup_import_event unset/false
        '''
        # NOTE: this should yield 0 created event _ one state event
        with patch.dict(vmadm.VMADM_STATE, MOCK_CLEAN_STATE), \
             patch.dict(vmadm.__salt__, {'vmadm.list': MagicMock(return_value=MOCK_VM_ONE)}):

            config = []

            ret = vmadm.validate(config)
            self.assertEqual(ret, (True, 'Valid beacon configuration'))

            ret = vmadm.beacon(config)
            res = [{'alias': 'vm1',
                    'tag': 'running/00000000-0000-0000-0000-000000000001',
                    'hostname': 'vm1',
                    'dns_domain': 'example.org'}]

            self.assertEqual(ret, res)

    def test_created(self):
        '''
        Test with one vm, create a 2nd one
        '''
        # NOTE: this should yield 1 created event + state event
        with patch.dict(vmadm.VMADM_STATE, MOCK_CLEAN_STATE), \
             patch.dict(vmadm.__salt__, {'vmadm.list': MagicMock(side_effect=[MOCK_VM_ONE, MOCK_VM_TWO_STARTED])}):

            config = []

            ret = vmadm.validate(config)
            self.assertEqual(ret, (True, 'Valid beacon configuration'))

            # Initial pass (Initialized state and do not yield created events at startup)
            ret = vmadm.beacon(config)

            # Second pass (After create a new vm)
            ret = vmadm.beacon(config)
            res = [{'alias': 'vm2',
                    'tag': 'created/00000000-0000-0000-0000-000000000002',
                    'hostname': 'vm2',
                    'dns_domain': 'example.org'},
                   {'alias': 'vm2',
                    'tag': 'running/00000000-0000-0000-0000-000000000002',
                    'hostname': 'vm2',
                    'dns_domain': 'example.org'}]

            self.assertEqual(ret, res)

    def test_deleted(self):
        '''
        Test with two vms and one gets destroyed
        '''
        # NOTE: this should yield 1 destroyed event
        with patch.dict(vmadm.VMADM_STATE, MOCK_CLEAN_STATE), \
             patch.dict(vmadm.__salt__, {'vmadm.list': MagicMock(side_effect=[MOCK_VM_TWO_STOPPED, MOCK_VM_ONE])}):

            config = []

            ret = vmadm.validate(config)
            self.assertEqual(ret, (True, 'Valid beacon configuration'))

            # Initial pass (Initialized state and do not yield created vms at startup)
            ret = vmadm.beacon(config)

            # Second pass (Destroying one vm)
            ret = vmadm.beacon(config)
            res = [{'alias': 'vm2',
                    'tag': 'deleted/00000000-0000-0000-0000-000000000002',
                    'hostname': 'vm2',
                    'dns_domain': 'example.org'}]

            self.assertEqual(ret, res)

    def test_complex(self):
        '''
        Test with two vms, stop one, delete one
        '''
        # NOTE: this should yield 1 delete and 2 import events
        with patch.dict(vmadm.VMADM_STATE, MOCK_CLEAN_STATE), \
             patch.dict(vmadm.__salt__, {'vmadm.list': MagicMock(side_effect=[MOCK_VM_TWO_STARTED, MOCK_VM_TWO_STOPPED, MOCK_VM_ONE])}):

            config = []

            ret = vmadm.validate(config)
            self.assertEqual(ret, (True, 'Valid beacon configuration'))

            # Initial pass (Initialized state and do not yield created events at startup)
            ret = vmadm.beacon(config)

            # Second pass (Stop one vm)
            ret = vmadm.beacon(config)
            res = [{'alias': 'vm2',
                    'tag': 'stopped/00000000-0000-0000-0000-000000000002',
                    'hostname': 'vm2',
                    'dns_domain': 'example.org'}]

            self.assertEqual(ret, res)

            # Third pass (Delete one vm)
            ret = vmadm.beacon(config)
            res = [{'alias': 'vm2',
                    'tag': 'deleted/00000000-0000-0000-0000-000000000002',
                    'hostname': 'vm2',
                    'dns_domain': 'example.org'}]

            self.assertEqual(ret, res)



# -*- coding: utf-8 -*-
'''
A module that adds data to the Pillar structure from a NetBox API.

.. versionadded:: 2019.2.0

Configuring the NetBox ext_pillar
---------------------------------

.. code-block:: yaml

  ext_pillar:
    - netbox:
        api_url: http://netbox_url.com/api/
        api_token: 123abc

Create a token in your NetBox instance at
http://netbox_url.com/user/api-tokens/

The following options are optional, and determine whether or not
the module will attempt to configure the ``proxy`` pillar data for
use with the napalm proxy-minion:

.. code-block:: yaml

  proxy_return: True
  proxy_username: admin

By default, this module will query the NetBox API for the platform
associated with the device, and use the 'NAPALM driver' field to
set the napalm proxy-minion driver. (Currently only 'napalm' is supported
for drivertype.)

This module currently only supports the napalm proxy minion and assumes
you will use SSH keys to authenticate to the network device.  If password
authentication is desired, it is recommended to create another ``proxy``
key in pillar_roots (or git_pillar) with just the ``passwd`` key and use
:py:func:`salt.renderers.gpg <salt.renderers.gpg>` to encrypt the value.
If any additional options for the proxy setup are needed they should also be
configured in pillar_roots.

Other available options:

site_details: ``True``
    Whether should retrieve details of the site the device belongs to.

site_prefixes: ``True``
    Whether should retrieve the prefixes of the site the device belongs to.
'''

from __future__ import absolute_import, print_function, unicode_literals
import logging

# Import Salt libs
import salt.utils.http
import salt.ext.ipaddress as ipaddress

log = logging.getLogger(__name__)


def ext_pillar(minion_id, pillar, *args, **kwargs):
    '''
    Query NetBox API for minion data
    '''
    if minion_id == '*':
        log.info('There\'s no data to collect from NetBox for the Master')
        return {}
    # Pull settings from kwargs
    api_url = kwargs['api_url'].rstrip('/')
    api_token = kwargs.get('api_token')
    site_details = kwargs.get('site_details', True)
    site_prefixes = kwargs.get('site_prefixes', True)
    proxy_username = kwargs.get('proxy_username', None)
    proxy_return = kwargs.get('proxy_return', True)
    ret = {}

    # Fetch device from API
    headers = {}
    if api_token:
        headers = {
            'Authorization': 'Token {}'.format(api_token)
        }
    device_url = '{api_url}/{app}/{endpoint}'.format(api_url=api_url,
                                                     app='dcim',
                                                     endpoint='devices')
    device_results = salt.utils.http.query(device_url,
                                           params={'name': minion_id},
                                           header_dict=headers,
                                           decode=True)
   # Check status code for API call
    if 'error' in device_results:
        log.error('API query failed for "%s", status code: %d',
                  minion_id, device_results['status'])
        log.error(device_results['error'])
        return ret
    # Assign results from API call to "netbox" key
    devices = device_results['dict']['results']
    if len(devices) == 1:
        ret['netbox'] = devices[0]
    elif len(devices) > 1:
        log.error('More than one device found for "%s"', minion_id)
        return ret
    else:
        log.error('Unable to pull NetBox data for "%s"', minion_id)
        return ret
    site_id = ret['netbox']['site']['id']
    site_name = ret['netbox']['site']['name']
    if site_details:
        log.debug('Retrieving site details for "%s" - site %s (ID %d)',
                  minion_id, site_name, site_id)
        site_url = '{api_url}/{app}/{endpoint}/{site_id}/'.format(api_url=api_url,
                                                                  app='dcim',
                                                                  endpoint='sites',
                                                                  site_id=site_id)
        site_details_ret = salt.utils.http.query(site_url,
                                                 header_dict=headers,
                                                 decode=True)
        if 'error' in site_details_ret:
            log.error('Unable to retrieve site details for %s (ID %d)',
                      site_name, site_id)
            log.error('Status code: %d, error: %s',
                      site_details_ret['status'],
                      site_details_ret['error'])
        else:
            ret['netbox']['site'] = site_details_ret['dict']
    if site_prefixes:
        log.debug('Retrieving site prefixes for "%s" - site %s (ID %d)',
                  minion_id, site_name, site_id)
        prefixes_url = '{api_url}/{app}/{endpoint}'.format(api_url=api_url,
                                                           app='ipam',
                                                           endpoint='prefixes')
        site_prefixes_ret = salt.utils.http.query(prefixes_url,
                                                  params={'site_id': site_id},
                                                  header_dict=headers,
                                                  decode=True)
        if 'error' in site_prefixes_ret:
            log.error('Unable to retrieve site prefixes for %s (ID %d)',
                      site_name, site_id)
            log.error('Status code: %d, error: %s',
                      site_prefixes_ret['status'],
                      site_prefixes_ret['error'])
        else:
            ret['netbox']['site']['prefixes'] = site_prefixes_ret['dict']['results']
    if proxy_return:
        # Attempt to add "proxy" key, based on platform API call
        try:
            # Fetch device from API
            platform_results = salt.utils.http.query(ret['netbox']['platform']['url'],
                                                     header_dict=headers,
                                                     decode=True)
            # Check status code for API call
            if 'error' in platform_results:
                log.info('API query failed for "%s": %s',
                         minion_id, platform_results['error'])
            # Assign results from API call to "proxy" key if the platform has a
            # napalm_driver defined.
            napalm_driver = platform_results['dict'].get('napalm_driver')
            if napalm_driver:
                ret['proxy'] = {
                    'host': str(ipaddress.IPv4Interface(
                                ret['netbox']['primary_ip4']['address']).ip),
                    'driver': napalm_driver,
                    'proxytype': 'napalm',
                }
                if proxy_username:
                    ret['proxy']['username'] = proxy_username

        except Exception:
            log.debug(
                'Could not create proxy config data for "%s"', minion_id)

    return ret



# -*- coding: utf-8 -*-
'''
This is the default compound matcher function.
'''
from __future__ import absolute_import, print_function, unicode_literals

import logging
from salt.ext import six  # pylint: disable=3rd-party-module-not-gated
import salt.loader
import salt.utils.minions  # pylint: disable=3rd-party-module-not-gated

HAS_RANGE = False
try:
    import seco.range  # pylint: disable=unused-import
    HAS_RANGE = True
except ImportError:
    pass

log = logging.getLogger(__name__)


def match(tgt):
    '''
    Runs the compound target check
    '''
    nodegroups = __opts__.get('nodegroups', {})
    matchers = salt.loader.matchers(__opts__)

    if not isinstance(tgt, six.string_types) and not isinstance(tgt, (list, tuple)):
        log.error('Compound target received that is neither string, list nor tuple')
        return False
    log.debug('compound_match: %s ? %s', __opts__['id'], tgt)
    ref = {'G': 'grain',
           'P': 'grain_pcre',
           'I': 'pillar',
           'J': 'pillar_pcre',
           'L': 'list',
           'N': None,      # Nodegroups should already be expanded
           'S': 'ipcidr',
           'E': 'pcre'}
    if HAS_RANGE:
        ref['R'] = 'range'

    results = []
    opers = ['and', 'or', 'not', '(', ')']

    if isinstance(tgt, six.string_types):
        words = tgt.split()
    else:
        # we make a shallow copy in order to not affect the passed in arg
        words = tgt[:]

    while words:
        word = words.pop(0)
        target_info = salt.utils.minions.parse_target(word)

        # Easy check first
        if word in opers:
            if results:
                if results[-1] == '(' and word in ('and', 'or'):
                    log.error('Invalid beginning operator after "(": %s', word)
                    return False
                if word == 'not':
                    if not results[-1] in ('and', 'or', '('):
                        results.append('and')
                results.append(word)
            else:
                # seq start with binary oper, fail
                if word not in ['(', 'not']:
                    log.error('Invalid beginning operator: %s', word)
                    return False
                results.append(word)

        elif target_info and target_info['engine']:
            if 'N' == target_info['engine']:
                # if we encounter a node group, just evaluate it in-place
                decomposed = salt.utils.minions.nodegroup_comp(target_info['pattern'], nodegroups)
                if decomposed:
                    words = decomposed + words
                continue

            engine = ref.get(target_info['engine'])
            if not engine:
                # If an unknown engine is called at any time, fail out
                log.error(
                    'Unrecognized target engine "%s" for target '
                    'expression "%s"', target_info['engine'], word
                )
                return False

            engine_args = [target_info['pattern']]
            engine_kwargs = {}
            if target_info['delimiter']:
                engine_kwargs['delimiter'] = target_info['delimiter']

            results.append(
                six.text_type(matchers['{0}_match.match'.format(engine)](*engine_args, **engine_kwargs))
            )

        else:
            # The match is not explicitly defined, evaluate it as a glob
            results.append(six.text_type(matchers['glob_match.match'](word)))

    results = ' '.join(results)
    log.debug('compound_match %s ? "%s" => "%s"', __opts__['id'], tgt, results)
    try:
        return eval(results)  # pylint: disable=W0123
    except Exception:
        log.error(
            'Invalid compound target: %s for results: %s', tgt, results)
        return False
    return False



# -*- coding: utf-8 -*-

# Import Salt Testing Libs
from __future__ import absolute_import
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase, skipIf
from tests.support.mock import (
    MagicMock,
    patch,
    NO_MOCK,
    NO_MOCK_REASON
)

# Import Salt libs
from salt.exceptions import SaltInvocationError
import salt.modules.selinux as selinux


@skipIf(NO_MOCK, NO_MOCK_REASON)
class SelinuxModuleTestCase(TestCase, LoaderModuleMockMixin):
    '''
    Test cases for salt.modules.selinux
    '''
    def setup_loader_modules(self):
        return {selinux: {}}

    def test_fcontext_get_policy_parsing(self):
        '''
        Test to verify that the parsing of the semanage output into fields is
        correct. Added with #45784.
        '''
        cases = [
            {
                'semanage_out': '/var/www(/.*)?     all files          system_u:object_r:httpd_sys_content_t:s0',
                'name': '/var/www(/.*)?',
                'filetype': 'all files',
                'sel_user': 'system_u',
                'sel_role': 'object_r',
                'sel_type': 'httpd_sys_content_t',
                'sel_level': 's0'
            },
            {
                'semanage_out': '/var/www(/.*)? all files          system_u:object_r:httpd_sys_content_t:s0',
                'name': '/var/www(/.*)?',
                'filetype': 'all files',
                'sel_user': 'system_u',
                'sel_role': 'object_r',
                'sel_type': 'httpd_sys_content_t',
                'sel_level': 's0'
            },
            {
                'semanage_out': '/var/lib/dhcp3?                                    directory          system_u:object_r:dhcp_state_t:s0',
                'name': '/var/lib/dhcp3?',
                'filetype': 'directory',
                'sel_user': 'system_u',
                'sel_role': 'object_r',
                'sel_type': 'dhcp_state_t',
                'sel_level': 's0'
            },
            {
                'semanage_out': '/var/lib/dhcp3?  directory          system_u:object_r:dhcp_state_t:s0',
                'name': '/var/lib/dhcp3?',
                'filetype': 'directory',
                'sel_user': 'system_u',
                'sel_role': 'object_r',
                'sel_type': 'dhcp_state_t',
                'sel_level': 's0'
            },
            {
                'semanage_out': '/var/lib/dhcp3? directory          system_u:object_r:dhcp_state_t:s0',
                'name': '/var/lib/dhcp3?',
                'filetype': 'directory',
                'sel_user': 'system_u',
                'sel_role': 'object_r',
                'sel_type': 'dhcp_state_t',
                'sel_level': 's0'
            }
        ]

        for case in cases:
            with patch.dict(selinux.__salt__, {'cmd.shell': MagicMock(return_value=case['semanage_out'])}):
                ret = selinux.fcontext_get_policy(case['name'])
                self.assertEqual(ret['filespec'], case['name'])
                self.assertEqual(ret['filetype'], case['filetype'])
                self.assertEqual(ret['sel_user'], case['sel_user'])
                self.assertEqual(ret['sel_role'], case['sel_role'])
                self.assertEqual(ret['sel_type'], case['sel_type'])
                self.assertEqual(ret['sel_level'], case['sel_level'])

    def test_parse_protocol_port_positive(self):
        '''
        Test to verify positive parsing name, protocol and port combinations
        '''
        cases = [
            {
                'name': 'tcp/80',
                'protocol': None,
                'port': None,
                'expected': ('tcp', '80')
            },
            {
                'name': 'udp/53',
                'protocol': None,
                'port': None,
                'expected': ('udp', '53')
            },
            {
                'name': 'tcp_test_dns',
                'protocol': 'tcp',
                'port': '53',
                'expected': ('tcp', '53')
            },
            {
                'name': 'udp_test/dns',
                'protocol': 'udp',
                'port': '53',
                'expected': ('udp', '53')
            },
        ]

        for case in cases:
            ret = selinux._parse_protocol_port(case['name'], case['protocol'], case['port'])
            self.assertTupleEqual(ret, case['expected'])

    def test_parse_protocol_port_negative(self):
        '''
        Test to verify negative parsing of name, protocol and port combinations
        '''
        cases = [
            {
                'name': 'invalid_name_no_args',
                'protocol': None,
                'port': None,
            },
            {
                'name': 'invalid_proto/80',
                'protocol': 'nottcp',
                'port': '80',
            },
            {
                'name': 'invalid_port',
                'protocol': 'tcp',
                'port': 'notaport',
            },
            {
                'name': 'missing_proto',
                'protocol': None,
                'port': '80',
            },
            {
                'name': 'missing_port',
                'protocol': 'udp',
                'port': None,
            },
        ]

        for case in cases:
            self.assertRaises(SaltInvocationError, selinux._parse_protocol_port, case['name'], case['protocol'],
                              case['port'])

    def test_port_get_policy_parsing(self):
        '''
        Test to verify that the parsing of the semanage port output into fields is correct.
        '''
        cases = [
            {
                'semanage_out': 'cma_port_t                     tcp      1050',
                'name': 'tcp/1050',
                'expected': {'sel_type': 'cma_port_t', 'protocol': 'tcp', 'port': '1050'},
            },
            {
                'semanage_out': 'cluster_port_t                 tcp      5149, 40040, 50006-50008',
                'name': 'tcp/40040',
                'expected': {'sel_type': 'cluster_port_t', 'protocol': 'tcp', 'port': '5149, 40040, 50006-50008'},
            },
            {
                'semanage_out': 'http_port_t                    tcp      9008, 8010, 9002-9003, 80, 81, 443, 488, 8008, 8009, 8443, 9000',
                'name': 'tcp/9000',
                'expected': {'sel_type': 'http_port_t', 'protocol': 'tcp', 'port': '9008, 8010, 9002-9003, 80, 81, 443, 488, 8008, 8009, 8443, 9000'},
            },
            {
                'semanage_out': 'vnc_port_t                     tcp      5985-5999, 5900-5983',
                'name': 'tcp/5985-5999',
                'expected': {'sel_type': 'vnc_port_t', 'protocol': 'tcp', 'port': '5985-5999, 5900-5983'},
            },
            {
                'semanage_out': 'zebra_port_t                   tcp      2606, 2608-2609, 2600-2604',
                'name': 'tcp/2608-2609',
                'expected': {'sel_type': 'zebra_port_t', 'protocol': 'tcp', 'port': '2606, 2608-2609, 2600-2604'},
            },
            {
                'semanage_out': 'radius_port_t                  udp      1645, 1812, 18120-18121',
                'name': 'tcp/18120-18121',
                'expected': {'sel_type': 'radius_port_t', 'protocol': 'udp', 'port': '1645, 1812, 18120-18121'},
            },
        ]

        for case in cases:
            with patch.dict(selinux.__salt__, {'cmd.shell': MagicMock(return_value=case['semanage_out'])}):
                ret = selinux.port_get_policy(case['name'])
                self.assertDictEqual(ret, case['expected'])



# -*- coding: utf-8 -*-
'''
SCP Module
==========

.. versionadded:: 2019.2.0

Module to copy files via `SCP <https://man.openbsd.org/scp>`_
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import python libs
import logging
import inspect

# Import salt modules
from salt.ext import six

try:
    import scp
    import paramiko
    HAS_SCP = True
except ImportError:
    HAS_SCP = False

__proxyenabled__ = ['*']
__virtualname__ = 'scp'

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_SCP:
        return False, 'Please install SCP for this modules: pip install scp'
    return __virtualname__


def _select_kwargs(**kwargs):
    paramiko_kwargs = {}
    scp_kwargs = {}
    PARAMIKO_KWARGS, _, _, _ = inspect.getargspec(paramiko.SSHClient.connect)
    PARAMIKO_KWARGS.pop(0)  # strip self
    PARAMIKO_KWARGS.append('auto_add_policy')
    SCP_KWARGS, _, _, _ = inspect.getargspec(scp.SCPClient.__init__)
    SCP_KWARGS.pop(0)  # strip self
    SCP_KWARGS.pop(0)  # strip transport arg (it is passed in _prepare_connection)
    for karg, warg in six.iteritems(kwargs):
        if karg in PARAMIKO_KWARGS and warg is not None:
            paramiko_kwargs[karg] = warg
        if karg in SCP_KWARGS and warg is not None:
            scp_kwargs[karg] = warg
    return paramiko_kwargs, scp_kwargs


def _prepare_connection(**kwargs):
    '''
    Prepare the underlying SSH connection with the remote target.
    '''
    paramiko_kwargs, scp_kwargs = _select_kwargs(**kwargs)
    ssh = paramiko.SSHClient()
    if paramiko_kwargs.pop('auto_add_policy', False):
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(**paramiko_kwargs)
    scp_client = scp.SCPClient(ssh.get_transport(),
                               **scp_kwargs)
    return scp_client


def get(remote_path,
        local_path='',
        recursive=False,
        preserve_times=False,
        **kwargs):
    '''
    Transfer files and directories from remote host to the localhost of the
    Minion.

    remote_path
        Path to retrieve from remote host. Since this is evaluated by scp on the
        remote host, shell wildcards and environment variables may be used.

    recursive: ``False``
        Transfer files and directories recursively.

    preserve_times: ``False``
        Preserve ``mtime`` and ``atime`` of transferred files and directories.

    hostname
        The hostname of the remote device.

    port: ``22``
        The port of the remote device.

    username
        The username required for SSH authentication on the device.

    password
        Used for password authentication. It is also used for private key
        decryption if ``passphrase`` is not given.

    passphrase
        Used for decrypting private keys.

    pkey
        An optional private key to use for authentication.

    key_filename
        The filename, or list of filenames, of optional private key(s) and/or
        certificates to try for authentication.

    timeout
        An optional timeout (in seconds) for the TCP connect.

    socket_timeout: ``10``
        The channel socket timeout in seconds.

    buff_size: ``16384``
        The size of the SCP send buffer.

    allow_agent: ``True``
        Set to ``False`` to disable connecting to the SSH agent.

    look_for_keys: ``True``
        Set to ``False`` to disable searching for discoverable private key
        files in ``~/.ssh/``

    banner_timeout
        An optional timeout (in seconds) to wait for the SSH banner to be
        presented.

    auth_timeout
        An optional timeout (in seconds) to wait for an authentication
        response.

    auto_add_policy: ``False``
        Automatically add the host to the ``known_hosts``.

    CLI Example:

    .. code-block:: bash

        salt '*' scp.get /var/tmp/file /tmp/file hostname=10.10.10.1 auto_add_policy=True
    '''
    scp_client = _prepare_connection(**kwargs)
    get_kwargs = {
        'recursive': recursive,
        'preserve_times': preserve_times
    }
    if local_path:
        get_kwargs['local_path'] = local_path
    return scp_client.get(remote_path, **get_kwargs)


def put(files,
        remote_path=None,
        recursive=False,
        preserve_times=False,
        saltenv='base',
        **kwargs):
    '''
    Transfer files and directories to remote host.

    files
        A single path or a list of paths to be transferred.

    remote_path
        The path on the remote device where to store the files.

    recursive: ``True``
        Transfer files and directories recursively.

    preserve_times: ``False``
        Preserve ``mtime`` and ``atime`` of transferred files and directories.

    hostname
        The hostname of the remote device.

    port: ``22``
        The port of the remote device.

    username
        The username required for SSH authentication on the device.

    password
        Used for password authentication. It is also used for private key
        decryption if ``passphrase`` is not given.

    passphrase
        Used for decrypting private keys.

    pkey
        An optional private key to use for authentication.

    key_filename
        The filename, or list of filenames, of optional private key(s) and/or
        certificates to try for authentication.

    timeout
        An optional timeout (in seconds) for the TCP connect.

    socket_timeout: ``10``
        The channel socket timeout in seconds.

    buff_size: ``16384``
        The size of the SCP send buffer.

    allow_agent: ``True``
        Set to ``False`` to disable connecting to the SSH agent.

    look_for_keys: ``True``
        Set to ``False`` to disable searching for discoverable private key
        files in ``~/.ssh/``

    banner_timeout
        An optional timeout (in seconds) to wait for the SSH banner to be
        presented.

    auth_timeout
        An optional timeout (in seconds) to wait for an authentication
        response.

    auto_add_policy: ``False``
        Automatically add the host to the ``known_hosts``.

    CLI Example:

    .. code-block:: bash

        salt '*' scp.put /path/to/file /var/tmp/file hostname=server1 auto_add_policy=True
    '''
    scp_client = _prepare_connection(**kwargs)
    put_kwargs = {
        'recursive': recursive,
        'preserve_times': preserve_times
    }
    if remote_path:
        put_kwargs['remote_path'] = remote_path
    cached_files = []
    if not isinstance(files, (list, tuple)):
        files = [files]
    for file_ in files:
        cached_file = __salt__['cp.cache_file'](file_, saltenv=saltenv)
        cached_files.append(cached_file)
    return scp_client.put(cached_files, **put_kwargs)



# -*- coding: utf-8 -*-
'''
This is the default pcre matcher.
'''
from __future__ import absolute_import, print_function, unicode_literals

import re


def match(tgt):
    '''
    Returns true if the passed pcre regex matches
    '''
    return bool(re.match(tgt, __opts__['id']))



# -*- coding: utf-8 -*-
'''
    :codeauthor: Eric Radman <ericshane@eradman.com>
'''

# Import Python libs
from __future__ import absolute_import

# Import Salt Testing libs
from tests.support.unit import skipIf, TestCase
from tests.support.mock import (
    MagicMock,
    patch,
    NO_MOCK,
    NO_MOCK_REASON,
)

# Import Salt libs
import salt.cloud

EXAMPLE_PROVIDERS = {
 'nyc_vcenter': {'vmware': {'driver': 'vmware',
                            'password': '123456',
                            'url': 'vca1.saltstack.com',
                            'minion': {
                                'master': 'providermaster',
                                'grains': {
                                    'providergrain': True
                                }
                            },
                            'profiles': {},
                            'user': 'root'}},
 'nj_vcenter': {'vmware': {'driver': 'vmware',
                           'password': '333',
                           'profiles': {},
                            'minion': {
                                'master': 'providermaster',
                                'grains': {
                                    'providergrain': True
                                }
                            },
                           'image': 'rhel6_64prod',
                           'url': 'vca2.saltstack.com',
                           'user': 'root'}}
}

EXAMPLE_PROFILES = {
 'nyc-vm': {'cluster': 'nycvirt',
            'datastore': 'datastore1',
            'devices': {'disk': {'Hard disk 1': {'controller': 'SCSI controller 1',
                                                 'size': 20}},
                        'network': {'Network Adapter 1': {'mac': '88:88:88:88:88:42',
                                                          'name': 'vlan50',
                                                          'switch_type': 'standard'}},
                        'scsi': {'SCSI controller 1': {'type': 'paravirtual'}}},
            'extra_config': {'mem.hotadd': 'yes'},
            'folder': 'coreinfra',
            'image': 'rhel6_64Guest',
            'minion': {
                'master': 'profilemaster',
                'grains':  {
                    'profilegrain': True
                }
            },
            'memory': '8GB',
            'num_cpus': 2,
            'power_on': True,
            'profile': 'nyc-vm',
            'provider': 'nyc_vcenter:vmware',
            'resourcepool': 'Resources'},
 'nj-vm': {'cluster': 'njvirt',
            'folder': 'coreinfra',
            'image': 'rhel6_64Guest',
            'memory': '8GB',
            'num_cpus': 2,
            'power_on': True,
            'profile': 'nj-vm',
            'provider': 'nj_vcenter:vmware',
            'resourcepool': 'Resources'},

}

EXAMPLE_MAP = {
 'nyc-vm': {'db1': {'cpus': 4,
                    'devices': {'disk': {'Hard disk 1': {'size': 40}},
                                'network': {'Network Adapter 1': {'mac': '22:4a:b2:92:b3:eb'}}},
                    'memory': '16GB',
                    'minion': {
                        'master': 'mapmaster',
                        'grains': {
                            'mapgrain': True
                        }
                    },
                    'name': 'db1'},
            'db2': {'name': 'db2',
                    'password': '456',
                    'provider': 'nj_vcenter:vmware'}},
 'nj-vm': {'db3': {'name': 'db3',
                   'password': '789',
                   }}
}


@skipIf(NO_MOCK, NO_MOCK_REASON)
class MapConfTest(TestCase):
    '''
    Validate evaluation of salt-cloud map configuration
    '''

    def test_cloud_map_merge_conf(self):
        '''
        Ensure that nested values can be selectivly overridden in a map file
        '''
        with patch('salt.config.check_driver_dependencies', MagicMock(return_value=True)), \
                patch('salt.cloud.Map.read', MagicMock(return_value=EXAMPLE_MAP)):
            self.maxDiff = None
            opts = {'extension_modules': '/var/cache/salt/master/extmods',
                    'optimization_order': [0, 1, 2],
                    'providers': EXAMPLE_PROVIDERS, 'profiles': EXAMPLE_PROFILES}
            cloud_map = salt.cloud.Map(opts)

            merged_profile = {
             'create': {'db1': {'cluster': 'nycvirt',
                                'cpus': 4,
                                'datastore': 'datastore1',
                                'devices': {'disk': {'Hard disk 1': {'controller': 'SCSI controller 1',
                                                                     'size': 40}},
                                            'network': {'Network Adapter 1': {'mac': '22:4a:b2:92:b3:eb',
                                                                              'name': 'vlan50',
                                                                              'switch_type': 'standard'}},
                                            'scsi': {'SCSI controller 1': {'type': 'paravirtual'}}},
                                'driver': 'vmware',
                                'extra_config': {'mem.hotadd': 'yes'},
                                'folder': 'coreinfra',
                                'image': 'rhel6_64Guest',
                                'memory': '16GB',
                                'minion': {'grains': {'mapgrain': True,
                                                      'profilegrain': True,
                                                      'providergrain': True},
                                'master': 'mapmaster'},
                                'name': 'db1',
                                'num_cpus': 2,
                                'password': '123456',
                                'power_on': True,
                                'profile': 'nyc-vm',
                                'provider': 'nyc_vcenter:vmware',
                                'resourcepool': 'Resources',
                                'url': 'vca1.saltstack.com',
                                'user': 'root'},
             'db2': {'cluster': 'nycvirt',
                     'datastore': 'datastore1',
                     'devices': {'disk': {'Hard disk 1': {'controller': 'SCSI controller 1',
                                                          'size': 20}},
                                 'network': {'Network Adapter 1': {'mac': '88:88:88:88:88:42',
                                                                   'name': 'vlan50',
                                                                   'switch_type': 'standard'}},
                                 'scsi': {'SCSI controller 1': {'type': 'paravirtual'}}},
                     'driver': 'vmware',
                     'extra_config': {'mem.hotadd': 'yes'},
                     'folder': 'coreinfra',
                     'image': 'rhel6_64Guest',
                     'memory': '8GB',
                     'minion': {'grains': {'profilegrain': True,
                                           'providergrain': True},
                                'master': 'profilemaster'},
                     'name': 'db2',
                     'num_cpus': 2,
                     'password': '456',
                     'power_on': True,
                     'profile': 'nyc-vm',
                     'provider': 'nj_vcenter:vmware',
                     'resourcepool': 'Resources',
                     'url': 'vca2.saltstack.com',
                     'user': 'root'},
             'db3': {'cluster': 'njvirt',
                     'driver': 'vmware',
                     'folder': 'coreinfra',
                     'image': 'rhel6_64Guest',
                     'memory': '8GB',
                     'minion': {'grains': {'providergrain': True},
                                'master': 'providermaster'},
                     'name': 'db3',
                     'num_cpus': 2,
                     'password': '789',
                     'power_on': True,
                     'profile': 'nj-vm',
                     'provider': 'nj_vcenter:vmware',
                     'resourcepool': 'Resources',
                     'url': 'vca2.saltstack.com',
                     'user': 'root'}}
            }

            # what we assert above w.r.t db2 using nj_vcenter:vmware provider:
            # - url is from the overriden nj_vcenter provider, not nyc_vcenter
            # - image from provider is still overridden by the nyc-vm profile
            # - password from map override is still overriding both the provider and profile password
            #
            # what we assert above about grain handling ( and provider/profile/map data in general )
            # - provider grains are able to be overridden by profile data
            # - provider grain sare overridden by map data
            # - profile data is overriden by map data
            # ie, the provider->profile->map inheritance works as expected
            map_data = cloud_map.map_data()
            self.assertEqual(map_data, merged_profile)



# -*- coding: utf-8 -*-
'''
Manage virt
===========

For the key certificate this state uses the external pillar in the master to call
for the generation and signing of certificates for systems running libvirt:

.. code-block:: yaml

    libvirt_keys:
      virt.keys
'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals
import fnmatch
import os

try:
    import libvirt  # pylint: disable=import-error
    HAS_LIBVIRT = True
except ImportError:
    HAS_LIBVIRT = False

# Import Salt libs
import salt.utils.args
import salt.utils.files
import salt.utils.stringutils
import salt.utils.versions
from salt.exceptions import CommandExecutionError

# Import 3rd-party libs
from salt.ext import six

__virtualname__ = 'virt'


def __virtual__():
    '''
    Only if virt module is available.

    :return:
    '''

    if 'virt.node_info' in __salt__:
        return __virtualname__
    return False


def keys(name, basepath='/etc/pki', **kwargs):
    '''
    Manage libvirt keys.

    name
        The name variable used to track the execution

    basepath
        Defaults to ``/etc/pki``, this is the root location used for libvirt
        keys on the hypervisor

    The following parameters are optional:

        country
            The country that the certificate should use.  Defaults to US.

        .. versionadded:: 2018.3.0

        state
            The state that the certificate should use.  Defaults to Utah.

        .. versionadded:: 2018.3.0

        locality
            The locality that the certificate should use.
            Defaults to Salt Lake City.

        .. versionadded:: 2018.3.0

        organization
            The organization that the certificate should use.
            Defaults to Salted.

        .. versionadded:: 2018.3.0

        expiration_days
            The number of days that the certificate should be valid for.
            Defaults to 365 days (1 year)

        .. versionadded:: 2018.3.0

    '''
    ret = {'name': name, 'changes': {}, 'result': True, 'comment': ''}

    # Grab all kwargs to make them available as pillar values
    # rename them to something hopefully unique to avoid
    # overriding anything existing
    pillar_kwargs = {}
    for key, value in six.iteritems(kwargs):
        pillar_kwargs['ext_pillar_virt.{0}'.format(key)] = value

    pillar = __salt__['pillar.ext']({'libvirt': '_'}, pillar_kwargs)
    paths = {
        'serverkey': os.path.join(basepath, 'libvirt',
                                  'private', 'serverkey.pem'),
        'servercert': os.path.join(basepath, 'libvirt',
                                   'servercert.pem'),
        'clientkey': os.path.join(basepath, 'libvirt',
                                  'private', 'clientkey.pem'),
        'clientcert': os.path.join(basepath, 'libvirt',
                                   'clientcert.pem'),
        'cacert': os.path.join(basepath, 'CA', 'cacert.pem')
    }

    for key in paths:
        p_key = 'libvirt.{0}.pem'.format(key)
        if p_key not in pillar:
            continue
        if not os.path.exists(os.path.dirname(paths[key])):
            os.makedirs(os.path.dirname(paths[key]))
        if os.path.isfile(paths[key]):
            with salt.utils.files.fopen(paths[key], 'r') as fp_:
                if salt.utils.stringutils.to_unicode(fp_.read()) != pillar[p_key]:
                    ret['changes'][key] = 'update'
        else:
            ret['changes'][key] = 'new'

    if not ret['changes']:
        ret['comment'] = 'All keys are correct'
    elif __opts__['test']:
        ret['result'] = None
        ret['comment'] = 'Libvirt keys are set to be updated'
        ret['changes'] = {}
    else:
        for key in ret['changes']:
            with salt.utils.files.fopen(paths[key], 'w+') as fp_:
                fp_.write(
                    salt.utils.stringutils.to_str(
                        pillar['libvirt.{0}.pem'.format(key)]
                    )
                )

        ret['comment'] = 'Updated libvirt certs and keys'

    return ret


def _virt_call(domain, function, section, comment,
               connection=None, username=None, password=None, **kwargs):
    '''
    Helper to call the virt functions. Wildcards supported.

    :param domain:
    :param function:
    :param section:
    :param comment:
    :return:
    '''
    ret = {'name': domain, 'changes': {}, 'result': True, 'comment': ''}
    targeted_domains = fnmatch.filter(__salt__['virt.list_domains'](), domain)
    changed_domains = list()
    ignored_domains = list()
    for targeted_domain in targeted_domains:
        try:
            response = __salt__['virt.{0}'.format(function)](targeted_domain,
                                                             connection=connection,
                                                             username=username,
                                                             password=password,
                                                             **kwargs)
            if isinstance(response, dict):
                response = response['name']
            changed_domains.append({'domain': targeted_domain, function: response})
        except libvirt.libvirtError as err:
            ignored_domains.append({'domain': targeted_domain, 'issue': six.text_type(err)})
    if not changed_domains:
        ret['result'] = False
        ret['comment'] = 'No changes had happened'
        if ignored_domains:
            ret['changes'] = {'ignored': ignored_domains}
    else:
        ret['changes'] = {section: changed_domains}
        ret['comment'] = comment

    return ret


def stopped(name, connection=None, username=None, password=None):
    '''
    Stops a VM by shutting it down nicely.

    .. versionadded:: 2016.3.0

    :param connection: libvirt connection URI, overriding defaults

        .. versionadded:: 2019.2.0
    :param username: username to connect with, overriding defaults

        .. versionadded:: 2019.2.0
    :param password: password to connect with, overriding defaults

        .. versionadded:: 2019.2.0

    .. code-block:: yaml

        domain_name:
          virt.stopped
    '''

    return _virt_call(name, 'shutdown', 'stopped', "Machine has been shut down",
                      connection=connection, username=username, password=password)


def powered_off(name, connection=None, username=None, password=None):
    '''
    Stops a VM by power off.

    .. versionadded:: 2016.3.0

    :param connection: libvirt connection URI, overriding defaults

        .. versionadded:: 2019.2.0
    :param username: username to connect with, overriding defaults

        .. versionadded:: 2019.2.0
    :param password: password to connect with, overriding defaults

        .. versionadded:: 2019.2.0

    .. code-block:: yaml

        domain_name:
          virt.stopped
    '''

    return _virt_call(name, 'stop', 'unpowered', 'Machine has been powered off',
                      connection=connection, username=username, password=password)


def running(name,
            cpu=None,
            mem=None,
            image=None,
            vm_type=None,
            disk_profile=None,
            disks=None,
            nic_profile=None,
            interfaces=None,
            graphics=None,
            seed=True,
            install=True,
            pub_key=None,
            priv_key=None,
            update=False,
            connection=None,
            username=None,
            password=None,
            os_type=None,
            arch=None):
    '''
    Starts an existing guest, or defines and starts a new VM with specified arguments.

    .. versionadded:: 2016.3.0

    :param name: name of the virtual machine to run
    :param cpu: number of CPUs for the virtual machine to create
    :param mem: amount of memory in MiB for the new virtual machine
    :param image: disk image to use for the first disk of the new VM

        .. deprecated:: 2019.2.0
    :param vm_type: force virtual machine type for the new VM. The default value is taken from
        the host capabilities. This could be useful for example to use ``'qemu'`` type instead
        of the ``'kvm'`` one.

        .. versionadded:: 2019.2.0
    :param disk_profile:
        Name of the disk profile to use for the new virtual machine

        .. versionadded:: 2019.2.0
    :param disks:
        List of disk to create for the new virtual machine.
        See :ref:`init-disk-def` for more details on the items on this list.

        .. versionadded:: 2019.2.0
    :param nic_profile:
        Name of the network interfaces profile to use for the new virtual machine

        .. versionadded:: 2019.2.0
    :param interfaces:
        List of network interfaces to create for the new virtual machine.
        See :ref:`init-nic-def` for more details on the items on this list.

        .. versionadded:: 2019.2.0
    :param graphics:
        Graphics device to create for the new virtual machine.
        See :ref:`init-graphics-def` for more details on this dictionary

        .. versionadded:: 2019.2.0
    :param saltenv:
        Fileserver environment (Default: ``'base'``).
        See :mod:`cp module for more details <salt.modules.cp>`

        .. versionadded:: 2019.2.0
    :param seed: ``True`` to seed the disk image. Only used when the ``image`` parameter is provided.
                 (Default: ``True``)

        .. versionadded:: 2019.2.0
    :param install: install salt minion if absent (Default: ``True``)

        .. versionadded:: 2019.2.0
    :param pub_key: public key to seed with (Default: ``None``)

        .. versionadded:: 2019.2.0
    :param priv_key: public key to seed with (Default: ``None``)

        .. versionadded:: 2019.2.0
    :param seed_cmd: Salt command to execute to seed the image. (Default: ``'seed.apply'``)

        .. versionadded:: 2019.2.0
    :param update: set to ``True`` to update a defined module. (Default: ``False``)

        .. versionadded:: 2019.2.0
    :param connection: libvirt connection URI, overriding defaults

        .. versionadded:: 2019.2.0
    :param username: username to connect with, overriding defaults

        .. versionadded:: 2019.2.0
    :param password: password to connect with, overriding defaults

        .. versionadded:: 2019.2.0
    :param os_type:
        type of virtualization as found in the ``//os/type`` element of the libvirt definition.
        The default value is taken from the host capabilities, with a preference for ``hvm``.
        Only used when creating a new virtual machine.

        .. versionadded:: Neon
    :param arch:
        architecture of the virtual machine. The default value is taken from the host capabilities,
        but ``x86_64`` is prefed over ``i686``. Only used when creating a new virtual machine.

        .. versionadded:: Neon

    .. rubric:: Example States

    Make sure an already-defined virtual machine called ``domain_name`` is running:

    .. code-block:: yaml

        domain_name:
          virt.running

    Do the same, but define the virtual machine if needed:

    .. code-block:: yaml

        domain_name:
          virt.running:
            - cpu: 2
            - mem: 2048
            - disk_profile: prod
            - disks:
              - name: system
                size: 8192
                overlay_image: True
                pool: default
                image: /path/to/image.qcow2
              - name: data
                size: 16834
            - nic_profile: prod
            - interfaces:
              - name: eth0
                mac: 01:23:45:67:89:AB
              - name: eth1
                type: network
                source: admin
            - graphics:
              - type: spice
                listen:
                  - type: address
                    address: 192.168.0.125

    '''

    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': '{0} is running'.format(name)
           }

    try:
        try:
            __salt__['virt.vm_state'](name)
            if __salt__['virt.vm_state'](name) != 'running':
                action_msg = 'started'
                if update:
                    status = __salt__['virt.update'](name,
                                                     cpu=cpu,
                                                     mem=mem,
                                                     disk_profile=disk_profile,
                                                     disks=disks,
                                                     nic_profile=nic_profile,
                                                     interfaces=interfaces,
                                                     graphics=graphics,
                                                     live=False,
                                                     connection=connection,
                                                     username=username,
                                                     password=password)
                    if status['definition']:
                        action_msg = 'updated and started'
                __salt__['virt.start'](name)
                ret['changes'][name] = 'Domain {0}'.format(action_msg)
                ret['comment'] = 'Domain {0} {1}'.format(name, action_msg)
            else:
                if update:
                    status = __salt__['virt.update'](name,
                                                     cpu=cpu,
                                                     mem=mem,
                                                     disk_profile=disk_profile,
                                                     disks=disks,
                                                     nic_profile=nic_profile,
                                                     interfaces=interfaces,
                                                     graphics=graphics,
                                                     connection=connection,
                                                     username=username,
                                                     password=password)
                    ret['changes'][name] = status
                    if status.get('errors', None):
                        ret['comment'] = 'Domain {0} updated, but some live update(s) failed'.format(name)
                    elif not status['definition']:
                        ret['comment'] = 'Domain {0} exists and is running'.format(name)
                    else:
                        ret['comment'] = 'Domain {0} updated, restart to fully apply the changes'.format(name)
                else:
                    ret['comment'] = 'Domain {0} exists and is running'.format(name)
        except CommandExecutionError:
            if image:
                salt.utils.versions.warn_until(
                    'Sodium',
                    '\'image\' parameter has been deprecated. Rather use the \'disks\' parameter '
                    'to override or define the image. \'image\' will be removed in {version}.'
                )
            __salt__['virt.init'](name,
                                  cpu=cpu,
                                  mem=mem,
                                  os_type=os_type,
                                  arch=arch,
                                  image=image,
                                  hypervisor=vm_type,
                                  disk=disk_profile,
                                  disks=disks,
                                  nic=nic_profile,
                                  interfaces=interfaces,
                                  graphics=graphics,
                                  seed=seed,
                                  install=install,
                                  pub_key=pub_key,
                                  priv_key=priv_key,
                                  connection=connection,
                                  username=username,
                                  password=password)
            ret['changes'][name] = 'Domain defined and started'
            ret['comment'] = 'Domain {0} defined and started'.format(name)
    except libvirt.libvirtError as err:
        # Something bad happened when starting / updating the VM, report it
        ret['comment'] = six.text_type(err)
        ret['result'] = False

    return ret


def snapshot(name, suffix=None, connection=None, username=None, password=None):
    '''
    Takes a snapshot of a particular VM or by a UNIX-style wildcard.

    .. versionadded:: 2016.3.0

    :param connection: libvirt connection URI, overriding defaults

        .. versionadded:: 2019.2.0
    :param username: username to connect with, overriding defaults

        .. versionadded:: 2019.2.0
    :param password: password to connect with, overriding defaults

        .. versionadded:: 2019.2.0

    .. code-block:: yaml

        domain_name:
          virt.snapshot:
            - suffix: periodic

        domain*:
          virt.snapshot:
            - suffix: periodic
    '''

    return _virt_call(name, 'snapshot', 'saved', 'Snapshot has been taken', suffix=suffix,
                      connection=connection, username=username, password=password)


# Deprecated states
def rebooted(name, connection=None, username=None, password=None):
    '''
    Reboots VMs

    .. versionadded:: 2016.3.0

    :param name:

    :param connection: libvirt connection URI, overriding defaults

        .. versionadded:: 2019.2.0
    :param username: username to connect with, overriding defaults

        .. versionadded:: 2019.2.0
    :param password: password to connect with, overriding defaults

        .. versionadded:: 2019.2.0
    '''

    return _virt_call(name, 'reboot', 'rebooted', "Machine has been rebooted",
                      connection=connection, username=username, password=password)


def unpowered(name):
    '''
    .. deprecated:: 2016.3.0
       Use :py:func:`~salt.modules.virt.powered_off` instead.

    Stops a VM by power off.

    .. versionadded:: 2016.3.0

    .. code-block:: yaml

        domain_name:
          virt.stopped
    '''

    return _virt_call(name, 'stop', 'unpowered', 'Machine has been powered off')


def saved(name, suffix=None):
    '''
    .. deprecated:: 2016.3.0
       Use :py:func:`~salt.modules.virt.snapshot` instead.

    Takes a snapshot of a particular VM or by a UNIX-style wildcard.

    .. versionadded:: 2016.3.0

    .. code-block:: yaml

        domain_name:
          virt.saved:
            - suffix: periodic

        domain*:
          virt.saved:
            - suffix: periodic
    '''

    return _virt_call(name, 'snapshot', 'saved', 'Snapshots has been taken', suffix=suffix)


def reverted(name, snapshot=None, cleanup=False):  # pylint: disable=redefined-outer-name
    '''
    .. deprecated:: 2016.3.0

    Reverts to the particular snapshot.

    .. versionadded:: 2016.3.0

    .. code-block:: yaml

        domain_name:
          virt.reverted:
            - cleanup: True

        domain_name_1:
          virt.reverted:
            - snapshot: snapshot_name
            - cleanup: False
    '''
    ret = {'name': name, 'changes': {}, 'result': False, 'comment': ''}

    try:
        domains = fnmatch.filter(__salt__['virt.list_domains'](), name)
        if not domains:
            ret['comment'] = 'No domains found for criteria "{0}"'.format(name)
        else:
            ignored_domains = list()
            if len(domains) > 1:
                ret['changes'] = {'reverted': list()}
            for domain in domains:
                result = {}
                try:
                    result = __salt__['virt.revert_snapshot'](domain, snapshot=snapshot, cleanup=cleanup)
                    result = {'domain': domain, 'current': result['reverted'], 'deleted': result['deleted']}
                except CommandExecutionError as err:
                    if len(domains) > 1:
                        ignored_domains.append({'domain': domain, 'issue': six.text_type(err)})
                if len(domains) > 1:
                    if result:
                        ret['changes']['reverted'].append(result)
                else:
                    ret['changes'] = result
                    break

            ret['result'] = len(domains) != len(ignored_domains)
            if ret['result']:
                ret['comment'] = 'Domain{0} has been reverted'.format(len(domains) > 1 and "s" or "")
            if ignored_domains:
                ret['changes']['ignored'] = ignored_domains
            if not ret['changes']['reverted']:
                ret['changes'].pop('reverted')
    except libvirt.libvirtError as err:
        ret['comment'] = six.text_type(err)
    except CommandExecutionError as err:
        ret['comment'] = six.text_type(err)

    return ret


def network_running(name,
                    bridge,
                    forward,
                    vport=None,
                    tag=None,
                    autostart=True,
                    connection=None,
                    username=None,
                    password=None):
    '''
    Defines and starts a new network with specified arguments.

    :param connection: libvirt connection URI, overriding defaults

        .. versionadded:: 2019.2.0
    :param username: username to connect with, overriding defaults

        .. versionadded:: 2019.2.0
    :param password: password to connect with, overriding defaults

        .. versionadded:: 2019.2.0

    .. code-block:: yaml

        domain_name:
          virt.network_define

    .. code-block:: yaml

        network_name:
          virt.network_define:
            - bridge: main
            - forward: bridge
            - vport: openvswitch
            - tag: 180
            - autostart: True

    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''
           }

    try:
        info = __salt__['virt.network_info'](name, connection=connection, username=username, password=password)
        if info:
            if info['active']:
                ret['comment'] = 'Network {0} exists and is running'.format(name)
            else:
                __salt__['virt.network_start'](name, connection=connection, username=username, password=password)
                ret['changes'][name] = 'Network started'
                ret['comment'] = 'Network {0} started'.format(name)
        else:
            __salt__['virt.network_define'](name,
                                            bridge,
                                            forward,
                                            vport,
                                            tag=tag,
                                            autostart=autostart,
                                            start=True,
                                            connection=connection,
                                            username=username,
                                            password=password)
            ret['changes'][name] = 'Network defined and started'
            ret['comment'] = 'Network {0} defined and started'.format(name)
    except libvirt.libvirtError as err:
        ret['result'] = False
        ret['comment'] = err.get_error_message()

    return ret


def pool_running(name,
                 ptype=None,
                 target=None,
                 permissions=None,
                 source=None,
                 transient=False,
                 autostart=True,
                 connection=None,
                 username=None,
                 password=None):
    '''
    Defines and starts a new pool with specified arguments.

    .. versionadded:: 2019.2.0

    :param ptype: libvirt pool type
    :param target: full path to the target device or folder. (Default: ``None``)
    :param permissions:
        target permissions. See :ref:`pool-define-permissions` for more details on this structure.
    :param source:
        dictionary containing keys matching the ``source_*`` parameters in function
        :func:`salt.modules.virt.pool_define`.
    :param transient:
        when set to ``True``, the pool will be automatically undefined after being stopped. (Default: ``False``)
    :param autostart:
        Whether to start the pool when booting the host. (Default: ``True``)
    :param start:
        When ``True``, define and start the pool, otherwise the pool will be left stopped.
    :param connection: libvirt connection URI, overriding defaults
    :param username: username to connect with, overriding defaults
    :param password: password to connect with, overriding defaults

    .. code-block:: yaml

        pool_name:
          virt.pool_define

    .. code-block:: yaml

        pool_name:
          virt.pool_define:
            - ptype: netfs
            - target: /mnt/cifs
            - permissions:
                - mode: 0770
                - owner: 1000
                - group: 100
            - source:
                - dir: samba_share
                - hosts:
                   one.example.com
                   two.example.com
                - format: cifs
            - autostart: True

    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''
           }

    try:
        info = __salt__['virt.pool_info'](name, connection=connection, username=username, password=password)
        if info:
            if info['state'] == 'running':
                ret['comment'] = 'Pool {0} exists and is running'.format(name)
            else:
                __salt__['virt.pool_start'](name, connection=connection, username=username, password=password)
                ret['changes'][name] = 'Pool started'
                ret['comment'] = 'Pool {0} started'.format(name)
        else:
            __salt__['virt.pool_define'](name,
                                         ptype=ptype,
                                         target=target,
                                         permissions=permissions,
                                         source_devices=(source or {}).get('devices', None),
                                         source_dir=(source or {}).get('dir', None),
                                         source_adapter=(source or {}).get('adapter', None),
                                         source_hosts=(source or {}).get('hosts', None),
                                         source_auth=(source or {}).get('auth', None),
                                         source_name=(source or {}).get('name', None),
                                         source_format=(source or {}).get('format', None),
                                         transient=transient,
                                         start=True,
                                         connection=connection,
                                         username=username,
                                         password=password)
            if autostart:
                __salt__['virt.pool_set_autostart'](name,
                                                    state='on' if autostart else 'off',
                                                    connection=connection,
                                                    username=username,
                                                    password=password)

            __salt__['virt.pool_build'](name,
                                        connection=connection,
                                        username=username,
                                        password=password)

            __salt__['virt.pool_start'](name,
                                        connection=connection,
                                        username=username,
                                        password=password)
            ret['changes'][name] = 'Pool defined and started'
            ret['comment'] = 'Pool {0} defined and started'.format(name)
    except libvirt.libvirtError as err:
        ret['comment'] = err.get_error_message()
        ret['result'] = False

    return ret



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.modules.pf as pf

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase
from tests.support.mock import (
    MagicMock,
    patch,
)


class PfTestCase(TestCase, LoaderModuleMockMixin):
    '''
    test modules.pf functions
    '''
    def setup_loader_modules(self):
        return {pf: {}}

    def test_enable_when_disabled(self):
        '''
        Tests enabling pf when it's not enabled yet.
        '''
        ret = {}
        ret['stderr'] = 'pf enabled'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(pf.enable()['changes'])

    def test_enable_when_enabled(self):
        '''
        Tests enabling pf when it already enabled.
        '''
        ret = {}
        ret['stderr'] = 'pfctl: pf already enabled'
        ret['retcode'] = 1
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertFalse(pf.enable()['changes'])

    def test_disable_when_enabled(self):
        '''
        Tests disabling pf when it's enabled.
        '''
        ret = {}
        ret['stderr'] = 'pf disabled'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(pf.disable()['changes'])

    def test_disable_when_disabled(self):
        '''
        Tests disabling pf when it already disabled.
        '''
        ret = {}
        ret['stderr'] = 'pfctl: pf not enabled'
        ret['retcode'] = 1
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertFalse(pf.disable()['changes'])

    def test_loglevel(self):
        '''
        Tests setting a loglevel.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            res = pf.loglevel('crit')
            mock_cmd.assert_called_once_with('pfctl -x crit',
                    output_loglevel='trace', python_shell=False)
            self.assertTrue(res['changes'])

    def test_load(self):
        '''
        Tests loading ruleset.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            res = pf.load()
            mock_cmd.assert_called_once_with(['pfctl', '-f', '/etc/pf.conf'],
                    output_loglevel='trace', python_shell=False)
            self.assertTrue(res['changes'])

    def test_load_noop(self):
        '''
        Tests evaluating but not actually loading ruleset.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            res = pf.load(noop=True)
            mock_cmd.assert_called_once_with(['pfctl', '-f', '/etc/pf.conf', '-n'],
                    output_loglevel='trace', python_shell=False)
            self.assertFalse(res['changes'])

    def test_flush(self):
        '''
        Tests a regular flush command.
        '''
        ret = {}
        ret['stderr'] = 'pf: statistics cleared'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            res = pf.flush('info')
            mock_cmd.assert_called_once_with('pfctl -v -F info',
                    output_loglevel='trace', python_shell=False)
            self.assertTrue(res['changes'])

    def test_flush_capital(self):
        '''
        Tests a flush command starting with a capital letter.
        '''
        ret = {}
        ret['stderr'] = '2 tables cleared'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            res = pf.flush('tables')
            mock_cmd.assert_called_once_with('pfctl -v -F Tables',
                    output_loglevel='trace', python_shell=False)
            self.assertTrue(res['changes'])

    def test_flush_without_changes(self):
        '''
        Tests a flush command that has no changes.
        '''
        ret = {}
        ret['stderr'] = '0 tables cleared'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertFalse(pf.flush('tables')['changes'])

    def test_table(self):
        '''
        Tests a regular table command.
        '''
        ret = {}
        ret['stderr'] = '42 addresses deleted'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(pf.table('flush', table='bad_hosts')['changes'])

    def test_table_expire(self):
        '''
        Tests the table expire command.
        '''
        ret = {}
        ret['stderr'] = '1/1 addresses expired.'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(pf.table('expire', table='bad_hosts', number=300)['changes'])

    def test_table_add_addresses(self):
        '''
        Tests adding addresses to a table.
        '''
        ret = {}
        ret['stderr'] = '2/2 addressess added.'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(pf.table('add', table='bad_hosts', addresses=['1.2.3.4', '5.6.7.8'])['changes'])

    def test_table_test_address(self):
        '''
        Tests testing addresses in a table.
        '''
        ret = {}
        ret['stderr'] = '1/2 addressess match.'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(pf.table('test', table='bad_hosts', addresses=['1.2.3.4'])['matches'])

    def test_table_no_changes(self):
        '''
        Tests a table command that has no changes.
        '''
        ret = {}
        ret['stderr'] = '0/1 addresses expired.'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertFalse(pf.table('expire', table='bad_hosts', number=300)['changes'])

    def test_table_show(self):
        '''
        Tests showing table contents.
        '''
        ret = {}
        ret['stdout'] = '1.2.3.4\n5.6.7.8'
        ret['retcode'] = 0
        expected = ['1.2.3.4', '5.6.7.8']
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertListEqual(pf.table('show', table='bad_hosts')['comment'], expected)

    def test_show(self):
        '''
        Tests a regular show command.
        '''
        ret = {}
        ret['stdout'] = 'block return\npass'
        ret['retcode'] = 0
        expected = ['block return', 'pass']
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertListEqual(pf.show('rules')['comment'], expected)

    def test_show_capital(self):
        '''
        Tests a show command starting with a capital letter.
        '''
        ret = {}
        ret['stdout'] = 'bad_hosts'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(pf.__salt__, {'cmd.run_all': mock_cmd}):
            res = pf.show('tables')
            mock_cmd.assert_called_once_with('pfctl -s Tables',
                    output_loglevel='trace', python_shell=False)
            self.assertFalse(res['changes'])



# -*- coding: utf-8 -*-
'''
unit tests for clustershell roster
'''
# Import Python libs
from __future__ import absolute_import

# Import Salt Testing libraries
from tests.support.unit import TestCase, skipIf
from tests.support.mock import (
        NO_MOCK,
        NO_MOCK_REASON,
        MagicMock,
        patch)

# Import third-party libs
try:
    from ClusterShell.NodeSet import NodeSet  # pylint: disable=unused-import
    HAS_CLUSTERSHELL = True
except (ImportError, OSError) as e:
    HAS_CLUSTERSHELL = False


@skipIf(NO_MOCK, NO_MOCK_REASON)
@skipIf(HAS_CLUSTERSHELL is False, 'Install Python Clustershell bindings before running these tests.')
class ClusterShellTestCase(TestCase):
    '''
    Test cases for clustershell roster
    '''
    def test_targets(self):
        mock_socket = MagicMock()
        mock_nodeset = MagicMock()
        mock_nodeset.NodeSet.return_value = ['foo']
        with patch.dict('sys.modules', **{'socket': mock_socket, 'ClusterShell.NodeSet': mock_nodeset}):
            import salt.roster.clustershell
            salt.roster.clustershell.__opts__ = {}
            with patch.dict(salt.roster.clustershell.__opts__, {'ssh_scan_ports': [1, 2, 3],
                'ssh_scan_timeout': 30}):
                # Reimports are necessary to re-init the namespace.
                import socket
                from ClusterShell.NodeSet import NodeSet
                ret = salt.roster.clustershell.targets('foo')
                mock_socket.gethostbyname.assert_any_call('foo')
                self.assertTrue('foo' in ret)
                self.assertTrue(ret['foo']['port'] == 3)



# -*- coding: utf-8 -*-
'''
.. versionadded:: 2017.7

Management of Zabbix Action object over Zabbix API.

:codeauthor: Jakub Sliva <jakub.sliva@ultimum.io>
'''
from __future__ import absolute_import
from __future__ import unicode_literals
import logging
import json

try:
    from salt.ext import six
    from salt.exceptions import SaltException
    IMPORTS_OK = True
except ImportError:
    IMPORTS_OK = False


log = logging.getLogger(__name__)


def __virtual__():
    '''
    Only make these states available if Zabbix module and run_query function is available
    and all 3rd party modules imported.
    '''
    if 'zabbix.run_query' in __salt__ and IMPORTS_OK:
        return True
    return False, 'Import zabbix or other needed modules failed.'


def present(name, params, **kwargs):
    '''
    Creates Zabbix Action object or if differs update it according defined parameters

    :param name: Zabbix Action name
    :param params: Definition of the Zabbix Action
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    If there is a need to get a value from current zabbix online (e.g. id of a hostgroup you want to put a discovered
    system into), put a dictionary with two keys "query_object" and "query_name" instead of the value.
    In this example we want to get object id of hostgroup named "Virtual machines" and "Databases".

    .. code-block:: yaml

        zabbix-action-present:
            zabbix_action.present:
                - name: VMs
                - params:
                    eventsource: 2
                    status: 0
                    filter:
                        evaltype: 2
                        conditions:
                            - conditiontype: 24
                              operator: 2
                              value: 'virtual'
                            - conditiontype: 24
                              operator: 2
                              value: 'kvm'
                    operations:
                        - operationtype: 2
                        - operationtype: 4
                          opgroup:
                              - groupid:
                                  query_object: hostgroup
                                  query_name: Virtual machines
                              - groupid:
                                  query_object: hostgroup
                                  query_name: Databases
    '''
    zabbix_id_mapper = __salt__['zabbix.get_zabbix_id_mapper']()

    dry_run = __opts__['test']
    ret = {'name': name, 'result': False, 'comment': '', 'changes': {}}

    # Create input params substituting functions with their results
    params['name'] = name
    params['operations'] = params['operations'] if 'operations' in params else []
    if 'filter' in params:
        params['filter']['conditions'] = params['filter']['conditions'] if 'conditions' in params['filter'] else []

    input_params = __salt__['zabbix.substitute_params'](params, **kwargs)
    log.info('Zabbix Action: input params: %s', six.text_type(json.dumps(input_params, indent=4)))

    search = {'output': 'extend',
              'selectOperations': 'extend',
              'selectFilter': 'extend',
              'filter': {
                  'name': name
              }}
    # GET Action object if exists
    action_get = __salt__['zabbix.run_query']('action.get', search, **kwargs)
    log.info('Zabbix Action: action.get result: %s', six.text_type(json.dumps(action_get, indent=4)))

    existing_obj = __salt__['zabbix.substitute_params'](action_get[0], **kwargs) \
        if action_get and len(action_get) == 1 else False

    if existing_obj:
        diff_params = __salt__['zabbix.compare_params'](input_params, existing_obj)
        log.info('Zabbix Action: input params: {%s', six.text_type(json.dumps(input_params, indent=4)))
        log.info('Zabbix Action: Object comparison result. Differences: %s', six.text_type(diff_params))

        if diff_params:
            diff_params[zabbix_id_mapper['action']] = existing_obj[zabbix_id_mapper['action']]
            # diff_params['name'] = 'VMs' - BUG - https://support.zabbix.com/browse/ZBX-12078
            log.info('Zabbix Action: update params: %s', six.text_type(json.dumps(diff_params, indent=4)))

            if dry_run:
                ret['result'] = True
                ret['comment'] = 'Zabbix Action "{0}" would be fixed.'.format(name)
                ret['changes'] = {name: {'old': 'Zabbix Action "{0}" differs '
                                                'in following parameters: {1}'.format(name, diff_params),
                                         'new': 'Zabbix Action "{0}" would correspond to definition.'.format(name)}}
            else:
                action_update = __salt__['zabbix.run_query']('action.update', diff_params, **kwargs)
                log.info('Zabbix Action: action.update result: %s', six.text_type(action_update))
                if action_update:
                    ret['result'] = True
                    ret['comment'] = 'Zabbix Action "{0}" updated.'.format(name)
                    ret['changes'] = {name: {'old': 'Zabbix Action "{0}" differed '
                                                    'in following parameters: {1}'.format(name, diff_params),
                                             'new': 'Zabbix Action "{0}" fixed.'.format(name)}}

        else:
            ret['result'] = True
            ret['comment'] = 'Zabbix Action "{0}" already exists and corresponds to a definition.'.format(name)

    else:
        if dry_run:
            ret['result'] = True
            ret['comment'] = 'Zabbix Action "{0}" would be created.'.format(name)
            ret['changes'] = {name: {'old': 'Zabbix Action "{0}" does not exist.'.format(name),
                                     'new': 'Zabbix Action "{0}" would be created according definition.'.format(name)}}
        else:
            # ACTION.CREATE
            action_create = __salt__['zabbix.run_query']('action.create', input_params, **kwargs)
            log.info('Zabbix Action: action.create result: ' + six.text_type(action_create))

            if action_create:
                ret['result'] = True
                ret['comment'] = 'Zabbix Action "{0}" created.'.format(name)
                ret['changes'] = {name: {'old': 'Zabbix Action "{0}" did not exist.'.format(name),
                                         'new': 'Zabbix Action "{0}" created according definition.'.format(name)}}

    return ret


def absent(name, **kwargs):
    '''
    Makes the Zabbix Action to be absent (either does not exist or delete it).

    :param name: Zabbix Action name
    :param _connection_user: Optional - zabbix user (can also be set in opts or pillar, see module's docstring)
    :param _connection_password: Optional - zabbix password (can also be set in opts or pillar, see module's docstring)
    :param _connection_url: Optional - url of zabbix frontend (can also be set in opts, pillar, see module's docstring)

    .. code-block:: yaml

        zabbix-action-absent:
            zabbix_action.absent:
                - name: Action name
    '''
    dry_run = __opts__['test']
    ret = {'name': name, 'result': False, 'comment': '', 'changes': {}}

    try:
        object_id = __salt__['zabbix.get_object_id_by_params']('action', {'filter': {'name': name}}, **kwargs)
    except SaltException:
        object_id = False

    if not object_id:
        ret['result'] = True
        ret['comment'] = 'Zabbix Action "{0}" does not exist.'.format(name)
    else:
        if dry_run:
            ret['result'] = True
            ret['comment'] = 'Zabbix Action "{0}" would be deleted.'.format(name)
            ret['changes'] = {name: {'old': 'Zabbix Action "{0}" exists.'.format(name),
                                     'new': 'Zabbix Action "{0}" would be deleted.'.format(name)}}
        else:
            action_delete = __salt__['zabbix.run_query']('action.delete', [object_id], **kwargs)

            if action_delete:
                ret['result'] = True
                ret['comment'] = 'Zabbix Action "{0}" deleted.'.format(name)
                ret['changes'] = {name: {'old': 'Zabbix Action "{0}" existed.'.format(name),
                                         'new': 'Zabbix Action "{0}" deleted.'.format(name)}}

    return ret



# coding=utf-8
'''
Test case for utils/__init__.py
'''
from __future__ import unicode_literals, print_function, absolute_import
from tests.support.unit import TestCase, skipIf
import salt.utils.environment

try:
    import pytest
except ImportError:
    pytest = None


@skipIf(pytest is None, 'PyTest is missing')
class UtilsTestCase(TestCase):
    '''
    Test case for utils/__init__.py
    '''
    def test_get_module_environment_empty(self):
        '''
        Test for salt.utils.get_module_environment
        Test if empty globals returns to an empty environment
        with the correct type.
        :return:
        '''
        out = salt.utils.environment.get_module_environment({})
        assert out == {}
        assert isinstance(out, dict)

    def test_get_module_environment_opts(self):
        '''
        Test for salt.utils.get_module_environment
        Test if __opts__ are visible.
        :return:
        '''
        expectation = {'message': 'Melting hard drives'}
        _globals = {
            '__opts__': {
                'system-environment': {
                    'modules': {
                        'system': {
                            '_': expectation,
                            },
                        },
                    },
                },
                '__file__': '/daemons/loose/modules/system.py',
            }
        assert salt.utils.environment.get_module_environment(_globals) == expectation

    def test_get_module_environment_pillars(self):
        '''
        Test for salt.utils.get_module_environment
        Test if __pillar__ is visible.
        :return:
        '''
        expectation = {'message': 'The CPU has shifted, and become decentralized.'}
        _globals = {
            '__opts__': {
                'system-environment': {
                    'electric': {
                        'interference': {
                            '_': expectation,
                            },
                        },
                    },
                },
                '__file__': '/piezo/electric/interference.py',
            }
        assert salt.utils.environment.get_module_environment(_globals) == expectation

    def test_get_module_environment_pillar_override(self):
        '''
        Test for salt.utils.get_module_environment
        Test if __pillar__ is overriding __opts__.
        :return:
        '''
        expectation = {'msg': 'The CPU has shifted, and become decentralized.'}
        _globals = {
            '__opts__': {
                'system-environment': {
                    'electric': {
                        'interference': {
                            '_': {'msg': 'Trololo!'},
                        },
                    },
                },
            },
            '__pillar__': {
                'system-environment': {
                    'electric': {
                        'interference': {
                            '_': expectation,
                            },
                        },
                    },
                },
            '__file__': '/piezo/electric/interference.py',
            }
        assert salt.utils.environment.get_module_environment(_globals) == expectation

    def test_get_module_environment_sname_found(self):
        '''
        Test for salt.utils.get_module_environment
        Section name and module name are found.
        :return:
        '''
        expectation = {'msg': 'All operators are on strike due to broken coffee machine!'}
        _globals = {
            '__opts__': {
                'system-environment': {
                    'jumping': {
                        'interference': {
                            '_': expectation,
                            },
                        },
                    },
                },
                '__file__': '/route/flapping/at_the_nap.py',
            }
        assert salt.utils.environment.get_module_environment(_globals) == {}

        _globals['__file__'] = '/route/jumping/interference.py'
        assert salt.utils.environment.get_module_environment(_globals) == expectation

    def test_get_module_environment_mname_found(self):
        '''
        Test for salt.utils.get_module_environment
        Module name is found.

        :return:
        '''
        expectation = {'msg': 'All operators are on strike due to broken coffee machine!'}
        _globals = {
            '__pillar__': {
                'system-environment': {
                    'jumping': {
                        'nonsense': {
                            '_': expectation,
                            },
                        },
                    },
                },
                '__file__': '/route/jumping/interference.py',
            }
        assert salt.utils.environment.get_module_environment(_globals) == {}
        _globals['__pillar__']['system-environment']['jumping']['interference'] = {}
        _globals['__pillar__']['system-environment']['jumping']['interference']['_'] = expectation
        assert salt.utils.environment.get_module_environment(_globals) == expectation

    def test_get_module_environment_vname_found(self):
        '''
        Test for salt.utils.get_module_environment
        Virtual name is found.

        :return:
        '''
        expectation = {'msg': 'All operators are on strike due to broken coffee machine!'}
        _globals = {
            '__virtualname__': 'nonsense',
            '__pillar__': {
                'system-environment': {
                    'jumping': {
                        'nonsense': {
                            '_': expectation,
                            },
                        },
                    },
                },
                '__file__': '/route/jumping/translation.py',
            }
        assert salt.utils.environment.get_module_environment(_globals) == expectation

    def test_get_module_environment_vname_overridden(self):
        '''
        Test for salt.utils.get_module_environment
        Virtual namespace overridden.

        :return:
        '''
        expectation = {'msg': 'New management.'}
        _globals = {
            '__virtualname__': 'nonsense',
            '__pillar__': {
                'system-environment': {
                    'funny': {
                        'translation': {
                            '_': expectation,
                        },
                        'nonsense': {
                            '_': {'msg': 'This is wrong'},
                            },
                        },
                    },
                },
                '__file__': '/lost/in/funny/translation.py',
            }

        assert salt.utils.environment.get_module_environment(_globals) == expectation



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Bo Maryniuk <bo@suse.de>`
'''
from __future__ import absolute_import, print_function, unicode_literals

import os
import sys
from tests.support.unit import TestCase, skipIf
from tests.support.mock import (
    NO_MOCK,
    NO_MOCK_REASON,
    MagicMock,
    patch)

import salt.exceptions
from salt.utils import thin
from salt.utils import json
import salt.utils.stringutils
from salt.utils.stringutils import to_bytes as bts
from salt.ext.six.moves import range

try:
    import pytest
except ImportError:
    pytest = None


@skipIf(NO_MOCK, NO_MOCK_REASON)
@skipIf(pytest is None, 'PyTest is missing')
class SSHThinTestCase(TestCase):
    '''
    TestCase for SaltSSH-related parts.
    '''
    def _popen(self, return_value=None, side_effect=None, returncode=0):
        '''
        Fake subprocess.Popen

        :return:
        '''

        proc = MagicMock()
        proc.communicate = MagicMock(return_value=return_value, side_effect=side_effect)
        proc.returncode = returncode
        popen = MagicMock(return_value=proc)

        return popen

    def _version_info(self, major=None, minor=None):
        '''
        Fake version info.

        :param major:
        :param minor:
        :return:
        '''
        class VersionInfo(tuple):
            pass

        vi = VersionInfo([major, minor])
        vi.major = major or sys.version_info.major
        vi.minor = minor or sys.version_info.minor

        return vi

    def _tarfile(self, getinfo=False):
        '''
        Fake tarfile handler.

        :return:
        '''
        spec = ['add', 'close']
        if getinfo:
            spec.append('getinfo')

        tf = MagicMock()
        tf.open = MagicMock(return_value=MagicMock(spec=spec))

        return tf

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.path.isfile', MagicMock(return_value=False))
    def test_get_ext_tops_cfg_missing_dependencies(self):
        '''
        Test thin.get_ext_tops contains all required dependencies.

        :return:
        '''
        cfg = {'namespace': {'py-version': [0, 0], 'path': '/foo', 'dependencies': []}}

        with pytest.raises(Exception) as err:
            thin.get_ext_tops(cfg)
        self.assertIn('Missing dependencies', str(err.value))
        self.assertTrue(thin.log.error.called)
        self.assertIn('Missing dependencies', thin.log.error.call_args[0][0])
        self.assertIn('jinja2, yaml, tornado, msgpack',
                      thin.log.error.call_args[0][0])

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.path.isfile', MagicMock(return_value=False))
    def test_get_ext_tops_cfg_missing_interpreter(self):
        '''
        Test thin.get_ext_tops contains interpreter configuration.

        :return:
        '''
        cfg = {'namespace': {'path': '/foo',
                             'dependencies': []}}
        with pytest.raises(salt.exceptions.SaltSystemExit) as err:
            thin.get_ext_tops(cfg)
        self.assertIn('missing specific locked Python version', str(err.value))

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.path.isfile', MagicMock(return_value=False))
    def test_get_ext_tops_cfg_wrong_interpreter(self):
        '''
        Test thin.get_ext_tops contains correct interpreter configuration.

        :return:
        '''
        cfg = {'namespace': {'path': '/foo',
                             'py-version': 2,
                             'dependencies': []}}

        with pytest.raises(salt.exceptions.SaltSystemExit) as err:
            thin.get_ext_tops(cfg)
        self.assertIn('specific locked Python version should be a list of '
                      'major/minor version', str(err.value))

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.path.isfile', MagicMock(return_value=False))
    def test_get_ext_tops_cfg_interpreter(self):
        '''
        Test thin.get_ext_tops interpreter configuration.

        :return:
        '''
        cfg = {'namespace': {'path': '/foo',
                             'py-version': [2, 6],
                             'dependencies': {'jinja2': '',
                                              'yaml': '',
                                              'tornado': '',
                                              'msgpack': ''}}}

        with pytest.raises(salt.exceptions.SaltSystemExit):
            thin.get_ext_tops(cfg)
        assert len(thin.log.warning.mock_calls) == 4
        assert sorted([x[1][1] for x in thin.log.warning.mock_calls]) == ['jinja2', 'msgpack', 'tornado', 'yaml']
        assert 'Module test has missing configuration' == thin.log.warning.mock_calls[0][1][0] % 'test'

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.path.isfile', MagicMock(return_value=False))
    def test_get_ext_tops_dependency_config_check(self):
        '''
        Test thin.get_ext_tops dependencies are importable

        :return:
        '''
        cfg = {'namespace': {'path': '/foo',
                             'py-version': [2, 6],
                             'dependencies': {'jinja2': '/jinja/foo.py',
                                              'yaml': '/yaml/',
                                              'tornado': '/tornado/wrong.rb',
                                              'msgpack': 'msgpack.sh'}}}

        with pytest.raises(salt.exceptions.SaltSystemExit) as err:
            thin.get_ext_tops(cfg)

        self.assertIn('Missing dependencies for the alternative version in the '
                      'external configuration', str(err.value))

        messages = {}
        for cl in thin.log.warning.mock_calls:
            messages[cl[1][1]] = cl[1][0] % (cl[1][1], cl[1][2])
        for mod in ['tornado', 'yaml', 'msgpack']:
            self.assertIn('not a Python importable module', messages[mod])
        self.assertIn('configured with not a file or does not exist',
                      messages['jinja2'])

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.path.isfile', MagicMock(return_value=True))
    def test_get_ext_tops_config_pass(self):
        '''
        Test thin.get_ext_tops configuration

        :return:
        '''
        cfg = {'namespace': {'path': '/foo',
                             'py-version': [2, 6],
                             'dependencies': {'jinja2': '/jinja/foo.py',
                                              'yaml': '/yaml/',
                                              'tornado': '/tornado/tornado.py',
                                              'msgpack': 'msgpack.py'}}}
        out = thin.get_ext_tops(cfg)
        assert out['namespace']['py-version'] == cfg['namespace']['py-version']
        assert out['namespace']['path'] == cfg['namespace']['path']
        assert sorted(out['namespace']['dependencies']) == sorted(['/tornado/tornado.py',
                                                                   '/jinja/foo.py', '/yaml/', 'msgpack.py'])

    @patch('salt.utils.thin.sys.argv', [None, '{"foo": "bar"}'])
    @patch('salt.utils.thin.get_tops', lambda **kw: kw)
    def test_gte(self):
        '''
        Test thin.gte external call for processing the info about tops per interpreter.

        :return:
        '''
        assert json.loads(thin.gte()).get('foo') == 'bar'

    def test_add_dep_path(self):
        '''
        Test thin._add_dependency function to setup dependency paths
        :return:
        '''
        container = []
        for pth in ['/foo/bar.py', '/something/else/__init__.py']:
            thin._add_dependency(container, type(str('obj'), (), {'__file__': pth})())
        assert '__init__' not in container[1]
        assert container == ['/foo/bar.py', '/something/else']

    def test_thin_path(self):
        '''
        Test thin.thin_path returns the expected path.

        :return:
        '''
        path = os.sep + os.path.join('path', 'to')
        expected = os.path.join(path, 'thin', 'thin.tgz')
        self.assertEqual(thin.thin_path(path), expected)

    def test_get_salt_call_script(self):
        '''
        Test get salt-call script rendered.

        :return:
        '''
        out = thin._get_salt_call('foo', 'bar', py26=[2, 6], py27=[2, 7], py34=[3, 4])
        for line in salt.utils.stringutils.to_str(out).split(os.linesep):
            if line.startswith('namespaces = {'):
                data = json.loads(line.replace('namespaces = ', '').strip())
                assert data.get('py26') == [2, 6]
                assert data.get('py27') == [2, 7]
                assert data.get('py34') == [3, 4]
            if line.startswith('syspaths = '):
                data = json.loads(line.replace('syspaths = ', ''))
                assert data == ['foo', 'bar']

    def test_get_ext_namespaces_empty(self):
        '''
        Test thin._get_ext_namespaces function returns an empty dictionary on nothing
        :return:
        '''
        for obj in [None, {}, []]:
            assert thin._get_ext_namespaces(obj) == {}

    def test_get_ext_namespaces(self):
        '''
        Test thin._get_ext_namespaces function returns namespaces properly out of the config.
        :return:
        '''
        cfg = {'ns': {'py-version': [2, 7]}}
        assert thin._get_ext_namespaces(cfg).get('ns') == (2, 7,)
        assert isinstance(thin._get_ext_namespaces(cfg).get('ns'), tuple)

    def test_get_ext_namespaces_failure(self):
        '''
        Test thin._get_ext_namespaces function raises an exception
        if python major/minor version is not configured.
        :return:
        '''
        with pytest.raises(salt.exceptions.SaltSystemExit):
            thin._get_ext_namespaces({'ns': {}})

    @patch('salt.utils.thin.salt', type(str('salt'), (), {'__file__': '/site-packages/salt'}))
    @patch('salt.utils.thin.jinja2', type(str('jinja2'), (), {'__file__': '/site-packages/jinja2'}))
    @patch('salt.utils.thin.yaml', type(str('yaml'), (), {'__file__': '/site-packages/yaml'}))
    @patch('salt.utils.thin.tornado', type(str('tornado'), (), {'__file__': '/site-packages/tornado'}))
    @patch('salt.utils.thin.msgpack', type(str('msgpack'), (), {'__file__': '/site-packages/msgpack'}))
    @patch('salt.utils.thin.certifi', type(str('certifi'), (), {'__file__': '/site-packages/certifi'}))
    @patch('salt.utils.thin.singledispatch', type(str('singledispatch'), (), {'__file__': '/site-packages/sdp'}))
    @patch('salt.utils.thin.singledispatch_helpers', type(str('singledispatch_helpers'), (), {'__file__': '/site-packages/sdp_hlp'}))
    @patch('salt.utils.thin.ssl_match_hostname', type(str('ssl_match_hostname'), (), {'__file__': '/site-packages/ssl_mh'}))
    @patch('salt.utils.thin.markupsafe', type(str('markupsafe'), (), {'__file__': '/site-packages/markupsafe'}))
    @patch('salt.utils.thin.backports_abc', type(str('backports_abc'), (), {'__file__': '/site-packages/backports_abc'}))
    @patch('salt.utils.thin.concurrent', type(str('concurrent'), (), {'__file__': '/site-packages/concurrent'}))
    @patch('salt.utils.thin.log', MagicMock())
    def test_get_tops(self):
        '''
        Test thin.get_tops to get top directories, based on the interpreter.
        :return:
        '''
        base_tops = ['/site-packages/salt', '/site-packages/jinja2', '/site-packages/yaml',
                     '/site-packages/tornado', '/site-packages/msgpack', '/site-packages/certifi',
                     '/site-packages/sdp', '/site-packages/sdp_hlp', '/site-packages/ssl_mh',
                     '/site-packages/markupsafe', '/site-packages/backports_abc', '/site-packages/concurrent']

        tops = thin.get_tops()
        assert len(tops) == len(base_tops)
        assert sorted(tops) == sorted(base_tops)

    @patch('salt.utils.thin.salt', type(str('salt'), (), {'__file__': '/site-packages/salt'}))
    @patch('salt.utils.thin.jinja2', type(str('jinja2'), (), {'__file__': '/site-packages/jinja2'}))
    @patch('salt.utils.thin.yaml', type(str('yaml'), (), {'__file__': '/site-packages/yaml'}))
    @patch('salt.utils.thin.tornado', type(str('tornado'), (), {'__file__': '/site-packages/tornado'}))
    @patch('salt.utils.thin.msgpack', type(str('msgpack'), (), {'__file__': '/site-packages/msgpack'}))
    @patch('salt.utils.thin.certifi', type(str('certifi'), (), {'__file__': '/site-packages/certifi'}))
    @patch('salt.utils.thin.singledispatch', type(str('singledispatch'), (), {'__file__': '/site-packages/sdp'}))
    @patch('salt.utils.thin.singledispatch_helpers', type(str('singledispatch_helpers'), (), {'__file__': '/site-packages/sdp_hlp'}))
    @patch('salt.utils.thin.ssl_match_hostname', type(str('ssl_match_hostname'), (), {'__file__': '/site-packages/ssl_mh'}))
    @patch('salt.utils.thin.markupsafe', type(str('markupsafe'), (), {'__file__': '/site-packages/markupsafe'}))
    @patch('salt.utils.thin.backports_abc', type(str('backports_abc'), (), {'__file__': '/site-packages/backports_abc'}))
    @patch('salt.utils.thin.concurrent', type(str('concurrent'), (), {'__file__': '/site-packages/concurrent'}))
    @patch('salt.utils.thin.log', MagicMock())
    def test_get_tops_extra_mods(self):
        '''
        Test thin.get_tops to get extra-modules alongside the top directories, based on the interpreter.
        :return:
        '''
        base_tops = ['/site-packages/salt',
                     '/site-packages/jinja2',
                     '/site-packages/yaml',
                     '/site-packages/tornado',
                     '/site-packages/msgpack',
                     '/site-packages/certifi',
                     '/site-packages/sdp',
                     '/site-packages/sdp_hlp',
                     '/site-packages/ssl_mh',
                     '/site-packages/concurrent',
                     '/site-packages/markupsafe',
                     '/site-packages/backports_abc',
                     os.sep + os.path.join('custom', 'foo'),
                     os.sep + os.path.join('custom', 'bar.py')]
        builtins = sys.version_info.major == 3 and 'builtins' or '__builtin__'
        foo = {'__file__': os.sep + os.path.join('custom', 'foo', '__init__.py')}
        bar = {'__file__': os.sep + os.path.join('custom', 'bar')}
        with patch('{}.__import__'.format(builtins),
                   MagicMock(side_effect=[type(str('foo'), (), foo),
                                          type(str('bar'), (), bar)])):
            tops = thin.get_tops(extra_mods='foo,bar')
        self.assertEqual(len(tops), len(base_tops))
        self.assertListEqual(sorted(tops), sorted(base_tops))

    @patch('salt.utils.thin.salt', type(str('salt'), (), {'__file__': '/site-packages/salt'}))
    @patch('salt.utils.thin.jinja2', type(str('jinja2'), (), {'__file__': '/site-packages/jinja2'}))
    @patch('salt.utils.thin.yaml', type(str('yaml'), (), {'__file__': '/site-packages/yaml'}))
    @patch('salt.utils.thin.tornado', type(str('tornado'), (), {'__file__': '/site-packages/tornado'}))
    @patch('salt.utils.thin.msgpack', type(str('msgpack'), (), {'__file__': '/site-packages/msgpack'}))
    @patch('salt.utils.thin.certifi', type(str('certifi'), (), {'__file__': '/site-packages/certifi'}))
    @patch('salt.utils.thin.singledispatch', type(str('singledispatch'), (), {'__file__': '/site-packages/sdp'}))
    @patch('salt.utils.thin.singledispatch_helpers', type(str('singledispatch_helpers'), (), {'__file__': '/site-packages/sdp_hlp'}))
    @patch('salt.utils.thin.ssl_match_hostname', type(str('ssl_match_hostname'), (), {'__file__': '/site-packages/ssl_mh'}))
    @patch('salt.utils.thin.markupsafe', type(str('markupsafe'), (), {'__file__': '/site-packages/markupsafe'}))
    @patch('salt.utils.thin.backports_abc', type(str('backports_abc'), (), {'__file__': '/site-packages/backports_abc'}))
    @patch('salt.utils.thin.concurrent', type(str('concurrent'), (), {'__file__': '/site-packages/concurrent'}))
    @patch('salt.utils.thin.log', MagicMock())
    def test_get_tops_so_mods(self):
        '''
        Test thin.get_tops to get extra-modules alongside the top directories, based on the interpreter.
        :return:
        '''
        base_tops = ['/site-packages/salt', '/site-packages/jinja2', '/site-packages/yaml',
                     '/site-packages/tornado', '/site-packages/msgpack', '/site-packages/certifi',
                     '/site-packages/sdp', '/site-packages/sdp_hlp', '/site-packages/ssl_mh', '/site-packages/concurrent',
                     '/site-packages/markupsafe', '/site-packages/backports_abc', '/custom/foo.so', '/custom/bar.so']
        builtins = sys.version_info.major == 3 and 'builtins' or '__builtin__'
        with patch('{}.__import__'.format(builtins),
                   MagicMock(side_effect=[type(str('salt'), (), {'__file__': '/custom/foo.so'}),
                                          type(str('salt'), (), {'__file__': '/custom/bar.so'})])):
            tops = thin.get_tops(so_mods='foo,bar')
        assert len(tops) == len(base_tops)
        assert sorted(tops) == sorted(base_tops)

    @patch('salt.utils.thin.gen_thin', MagicMock(return_value='/path/to/thin/thin.tgz'))
    @patch('salt.utils.hashutils.get_hash', MagicMock(return_value=12345))
    def test_thin_sum(self):
        '''
        Test thin.thin_sum function.

        :return:
        '''
        assert thin.thin_sum('/cachedir', form='sha256')[1] == 12345
        thin.salt.utils.hashutils.get_hash.assert_called()
        assert thin.salt.utils.hashutils.get_hash.call_count == 1

        path, form = thin.salt.utils.hashutils.get_hash.call_args[0]
        assert path == '/path/to/thin/thin.tgz'
        assert form == 'sha256'

    @patch('salt.utils.thin.gen_min', MagicMock(return_value='/path/to/thin/min.tgz'))
    @patch('salt.utils.hashutils.get_hash', MagicMock(return_value=12345))
    def test_min_sum(self):
        '''
        Test thin.thin_sum function.

        :return:
        '''
        assert thin.min_sum('/cachedir', form='sha256') == 12345
        thin.salt.utils.hashutils.get_hash.assert_called()
        assert thin.salt.utils.hashutils.get_hash.call_count == 1

        path, form = thin.salt.utils.hashutils.get_hash.call_args[0]
        assert path == '/path/to/thin/min.tgz'
        assert form == 'sha256'

    @patch('salt.utils.thin.sys.version_info', (2, 5))
    @patch('salt.exceptions.SaltSystemExit', Exception)
    def test_gen_thin_fails_ancient_python_version(self):
        '''
        Test thin.gen_thin function raises an exception
        if Python major/minor version is lower than 2.6

        :return:
        '''
        with pytest.raises(salt.exceptions.SaltSystemExit) as err:
            thin.sys.exc_clear = lambda: None
            thin.gen_thin('')
        self.assertIn('The minimum required python version to run salt-ssh is '
                      '"2.6"', str(err.value))

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.makedirs', MagicMock())
    @patch('salt.utils.files.fopen', MagicMock())
    @patch('salt.utils.thin._get_salt_call', MagicMock())
    @patch('salt.utils.thin._get_ext_namespaces', MagicMock())
    @patch('salt.utils.thin.get_tops', MagicMock(return_value=['/foo3', '/bar3']))
    @patch('salt.utils.thin.get_ext_tops', MagicMock(return_value={}))
    @patch('salt.utils.thin.os.path.isfile', MagicMock())
    @patch('salt.utils.thin.os.path.isdir', MagicMock(return_value=True))
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.remove', MagicMock())
    @patch('salt.utils.thin.os.path.exists', MagicMock())
    @patch('salt.utils.path.os_walk', MagicMock(return_value=[]))
    @patch('salt.utils.thin.subprocess.Popen',
           _popen(None, side_effect=[(bts('2.7'), bts('')), (bts('["/foo27", "/bar27"]'), bts(''))]))
    @patch('salt.utils.thin.tarfile', MagicMock())
    @patch('salt.utils.thin.zipfile', MagicMock())
    @patch('salt.utils.thin.os.getcwd', MagicMock())
    @patch('salt.utils.thin.os.chdir', MagicMock())
    @patch('salt.utils.thin.tempfile', MagicMock())
    @patch('salt.utils.thin.shutil', MagicMock())
    @patch('salt.utils.thin._six.PY3', True)
    @patch('salt.utils.thin._six.PY2', False)
    @patch('salt.utils.thin.sys.version_info', _version_info(None, 3, 6))
    def test_gen_thin_compression_fallback_py3(self):
        '''
        Test thin.gen_thin function if fallbacks to the gzip compression, once setup wrong.
        NOTE: Py2 version of this test is not required, as code shares the same spot across the versions.

        :return:
        '''
        thin.gen_thin('', compress='arj')
        thin.log.warning.assert_called()
        pt, msg = thin.log.warning.mock_calls[0][1]
        assert pt % msg == 'Unknown compression type: "arj". Falling back to "gzip" compression.'
        thin.zipfile.ZipFile.assert_not_called()
        thin.tarfile.open.assert_called()

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.makedirs', MagicMock())
    @patch('salt.utils.files.fopen', MagicMock())
    @patch('salt.utils.thin._get_salt_call', MagicMock())
    @patch('salt.utils.thin._get_ext_namespaces', MagicMock())
    @patch('salt.utils.thin.get_tops', MagicMock(return_value=['/foo3', '/bar3']))
    @patch('salt.utils.thin.get_ext_tops', MagicMock(return_value={}))
    @patch('salt.utils.thin.os.path.isfile', MagicMock())
    @patch('salt.utils.thin.os.path.isdir', MagicMock(return_value=False))
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.remove', MagicMock())
    @patch('salt.utils.thin.os.path.exists', MagicMock())
    @patch('salt.utils.path.os_walk', MagicMock(return_value=[]))
    @patch('salt.utils.thin.subprocess.Popen',
           _popen(None, side_effect=[(bts('2.7'), bts('')), (bts('["/foo27", "/bar27"]'), bts(''))]))
    @patch('salt.utils.thin.tarfile', MagicMock())
    @patch('salt.utils.thin.zipfile', MagicMock())
    @patch('salt.utils.thin.os.getcwd', MagicMock())
    @patch('salt.utils.thin.os.chdir', MagicMock())
    @patch('salt.utils.thin.tempfile', MagicMock(mkdtemp=MagicMock(return_value='')))
    @patch('salt.utils.thin.shutil', MagicMock())
    @patch('salt.utils.thin._six.PY3', True)
    @patch('salt.utils.thin._six.PY2', False)
    @patch('salt.utils.thin.sys.version_info', _version_info(None, 3, 6))
    def test_gen_thin_control_files_written_py3(self):
        '''
        Test thin.gen_thin function if control files are written (version, salt-call etc).
        NOTE: Py2 version of this test is not required, as code shares the same spot across the versions.

        :return:
        '''
        thin.gen_thin('')
        arc_name, arc_mode = thin.tarfile.method_calls[0][1]
        self.assertEqual(arc_name, os.path.join('thin', 'thin.tgz'))
        self.assertEqual(arc_mode, 'w:gz')
        for idx, fname in enumerate(['version', '.thin-gen-py-version', 'salt-call', 'supported-versions']):
            name = thin.tarfile.open().method_calls[idx + 4][1][0]
            self.assertEqual(name, fname)
        thin.tarfile.open().close.assert_called()

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.makedirs', MagicMock())
    @patch('salt.utils.files.fopen', MagicMock())
    @patch('salt.utils.thin._get_salt_call', MagicMock())
    @patch('salt.utils.thin._get_ext_namespaces', MagicMock())
    @patch('salt.utils.thin.get_tops', MagicMock(return_value=['/salt', '/bar3']))
    @patch('salt.utils.thin.get_ext_tops', MagicMock(return_value={}))
    @patch('salt.utils.thin.os.path.isfile', MagicMock())
    @patch('salt.utils.thin.os.path.isdir', MagicMock(return_value=True))
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.remove', MagicMock())
    @patch('salt.utils.thin.os.path.exists', MagicMock())
    @patch('salt.utils.path.os_walk',
           MagicMock(return_value=(('root', [], ['r1', 'r2', 'r3']), ('root2', [], ['r4', 'r5', 'r6']))))
    @patch('salt.utils.thin.subprocess.Popen',
           _popen(None, side_effect=[(bts('2.7'), bts('')), (bts('["/foo27", "/bar27"]'), bts(''))]))
    @patch('salt.utils.thin.tarfile', _tarfile(None))
    @patch('salt.utils.thin.zipfile', MagicMock())
    @patch('salt.utils.thin.os.getcwd', MagicMock())
    @patch('salt.utils.thin.os.chdir', MagicMock())
    @patch('salt.utils.thin.tempfile', MagicMock())
    @patch('salt.utils.thin.shutil', MagicMock())
    @patch('salt.utils.thin._six.PY3', True)
    @patch('salt.utils.thin._six.PY2', False)
    @patch('salt.utils.thin.sys.version_info', _version_info(None, 3, 6))
    @patch('salt.utils.hashutils.DigestCollector', MagicMock())
    def test_gen_thin_main_content_files_written_py3(self):
        '''
        Test thin.gen_thin function if main content files are written.
        NOTE: Py2 version of this test is not required, as code shares the same spot across the versions.

        :return:
        '''
        thin.gen_thin('')
        files = []
        for py in ('py2', 'py2', 'py3', 'pyall'):
            for i in range(1, 4):
                files.append(os.path.join(py, 'root', 'r{0}'.format(i)))
            for i in range(4, 7):
                files.append(os.path.join(py, 'root2', 'r{0}'.format(i)))
        for cl in thin.tarfile.open().method_calls[:-6]:
            arcname = cl[2].get('arcname')
            self.assertIn(arcname, files)
            files.pop(files.index(arcname))
        self.assertFalse(files)

    @patch('salt.exceptions.SaltSystemExit', Exception)
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.makedirs', MagicMock())
    @patch('salt.utils.files.fopen', MagicMock())
    @patch('salt.utils.thin._get_salt_call', MagicMock())
    @patch('salt.utils.thin._get_ext_namespaces', MagicMock())
    @patch('salt.utils.thin.get_tops', MagicMock(return_value=[]))
    @patch('salt.utils.thin.get_ext_tops',
           MagicMock(return_value={'namespace': {'py-version': [2, 7],
                                                 'path': '/opt/2015.8/salt',
                                                 'dependencies': ['/opt/certifi', '/opt/whatever']}}))
    @patch('salt.utils.thin.os.path.isfile', MagicMock())
    @patch('salt.utils.thin.os.path.isdir', MagicMock(return_value=True))
    @patch('salt.utils.thin.log', MagicMock())
    @patch('salt.utils.thin.os.remove', MagicMock())
    @patch('salt.utils.thin.os.path.exists', MagicMock())
    @patch('salt.utils.path.os_walk',
           MagicMock(return_value=(('root', [], ['r1', 'r2', 'r3']), ('root2', [], ['r4', 'r5', 'r6']))))
    @patch('salt.utils.thin.subprocess.Popen',
           _popen(None, side_effect=[(bts('2.7'), bts('')), (bts('["/foo27", "/bar27"]'), bts(''))]))
    @patch('salt.utils.thin.tarfile', _tarfile(None))
    @patch('salt.utils.thin.zipfile', MagicMock())
    @patch('salt.utils.thin.os.getcwd', MagicMock())
    @patch('salt.utils.thin.os.chdir', MagicMock())
    @patch('salt.utils.thin.tempfile', MagicMock(mkdtemp=MagicMock(return_value='')))
    @patch('salt.utils.thin.shutil', MagicMock())
    @patch('salt.utils.thin._six.PY3', True)
    @patch('salt.utils.thin._six.PY2', False)
    @patch('salt.utils.thin.sys.version_info', _version_info(None, 3, 6))
    @patch('salt.utils.hashutils.DigestCollector', MagicMock())
    def test_gen_thin_ext_alternative_content_files_written_py3(self):
        '''
        Test thin.gen_thin function if external alternative content files are written.
        NOTE: Py2 version of this test is not required, as code shares the same spot across the versions.

        :return:
        '''
        thin.gen_thin('')
        files = []
        for py in ('pyall', 'pyall', 'py2'):
            for i in range(1, 4):
                files.append(
                    os.path.join('namespace', py, 'root', 'r{0}'.format(i)))
            for i in range(4, 7):
                files.append(
                    os.path.join('namespace', py, 'root2', 'r{0}'.format(i)))

        for idx, cl in enumerate(thin.tarfile.open().method_calls[12:-6]):
            arcname = cl[2].get('arcname')
            self.assertIn(arcname, files)
            files.pop(files.index(arcname))
        self.assertFalse(files)

    def test_get_supported_py_config_typecheck(self):
        '''
        Test collecting proper py-versions. Should return bytes type.
        :return:
        '''
        tops = {}
        ext_cfg = {}
        out = thin._get_supported_py_config(tops=tops, extended_cfg=ext_cfg)
        assert type(salt.utils.stringutils.to_bytes('')) == type(out)

    def test_get_supported_py_config_base_tops(self):
        '''
        Test collecting proper py-versions. Should return proper base tops.
        :return:
        '''
        tops = {'3': ['/groundkeepers', '/stole'], '2': ['/the-root', '/password']}
        ext_cfg = {}
        out = salt.utils.stringutils.to_str(thin._get_supported_py_config(
            tops=tops, extended_cfg=ext_cfg)).strip().split(os.linesep)
        self.assertEqual(len(out), 2)
        for t_line in ['py3:3:0', 'py2:2:7']:
            self.assertIn(t_line, out)

    def test_get_supported_py_config_ext_tops(self):
        '''
        Test collecting proper py-versions. Should return proper ext conf tops.
        :return:
        '''
        tops = {}
        ext_cfg = {
            'solar-interference': {
                'py-version': [2, 6]},
            'second-system-effect': {
                'py-version': [2, 7]}}
        out = salt.utils.stringutils.to_str(thin._get_supported_py_config(
            tops=tops, extended_cfg=ext_cfg)).strip().split(os.linesep)
        for t_line in ['second-system-effect:2:7', 'solar-interference:2:6']:
            self.assertIn(t_line, out)



# -*- coding: utf-8 -*-
'''
Integration tests for the lxd states
'''
# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Testing libs
from tests.support.helpers import flaky

# Import Lxd Test Case
import tests.integration.states.test_lxd


@flaky
class LxdContainerTestCase(tests.integration.states.test_lxd.LxdTestCase):

    def setUp(self):
        self.run_state(
            'lxd_image.present',
            name='images:centos/7',
            source={
                'name': 'centos/7',
                'type': 'simplestreams',
                'server': 'https://images.linuxcontainers.org',
            },
        )

    def tearDown(self):
        self.run_state('lxd_image.absent', name='images:centos/7',)
        self.run_state('lxd_container.absent', name='test-container',)

    def test_02__create_container(self):
        ret = self.run_state(
            'lxd_container.present',
            name='test-container',
            running=True,
            source={'type': 'image', 'alias': 'images:centos/7'}
        )
        name = 'lxd_container_|-test-container_|-test-container_|-present'
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]['changes']['started'] == 'Started the container "test-container"'

    def test_03__change_container(self):
        self.run_state(
            'lxd_container.present',
            name='test-container',
            running=True,
            source={'type': 'image', 'alias': 'images:centos/7'}
        )
        ret = self.run_state(
            'lxd_container.present',
            name='test-container',
            running=True,
            source={'type': 'image', 'alias': 'images:centos/7'},
            restart_on_change=True,
            config=[
                {'key': 'boot.autostart', 'value': 1},
                {'key': 'security.privileged', 'value': '1'},
            ],
        )
        name = 'lxd_container_|-test-container_|-test-container_|-present'
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]['changes']['config'] == {
            'boot.autostart': 'Added config key "boot.autostart" = "1"',
            'security.privileged': 'Added config key "security.privileged" = "1"',
        }

    def test_08__running_container(self):
        self.run_state(
            'lxd_container.present',
            name='test-container',
            running=True,
            source={'type': 'image', 'alias': 'images:centos/7'}
        )
        ret = self.run_state(
            'lxd_container.running',
            name='test-container',
        )
        self.assertSaltTrueReturn(ret)
        name = 'lxd_container_|-test-container_|-test-container_|-running'
        assert name in ret
        assert not ret[name]['changes']
        assert ret[name]['comment'] == 'The container "test-container" is already running'
        ret = self.run_state(
            'lxd_container.running',
            name='test-container',
            restart=True,
        )
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]['changes'] == {'restarted': 'Restarted the container "test-container"'}
        assert ret[name]['comment'] == 'Restarted the container "test-container"'

    def test_09__stop_container(self):
        self.run_state(
            'lxd_container.present',
            name='test-container',
            running=True,
            source={'type': 'image', 'alias': 'images:centos/7'}
        )
        ret = self.run_state(
            'lxd_container.stopped',
            name='test-container',
        )
        name = 'lxd_container_|-test-container_|-test-container_|-stopped'
        self.assertSaltTrueReturn(ret)
        assert ret[name]['changes'] == {'stopped': 'Stopped the container "test-container"'}
        ret = self.run_state(
            'lxd_container.stopped',
            name='test-container',
        )
        name = 'lxd_container_|-test-container_|-test-container_|-stopped'
        self.assertSaltTrueReturn(ret)
        assert not ret[name]['changes']

    def test_10__delete_container(self):
        self.run_state(
            'lxd_container.present',
            name='test-container',
            running=True,
            source={'type': 'image', 'alias': 'images:centos/7'}
        )
        ret = self.run_state(
            'lxd_container.absent',
            name='test-container',
        )
        name = 'lxd_container_|-test-container_|-test-container_|-absent'
        assert name in ret
        assert ret[name]['result'] is False
        ret = self.run_state(
            'lxd_container.stopped',
            name='test-container',
        )
        name = 'lxd_container_|-test-container_|-test-container_|-stopped'
        assert name in ret
        assert ret[name]['result'] is True
        ret = self.run_state(
            'lxd_container.absent',
            name='test-container',
        )
        name = 'lxd_container_|-test-container_|-test-container_|-absent'
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]['changes'] == {'deleted': 'Container "test-container" has been deleted.'}



# -*- coding: utf-8 -*-
'''
Beacon to fire events at failed login of users

.. versionadded:: 2015.5.0

Example Configuration
=====================

.. code-block:: yaml

    # Fire events on all failed logins
    beacons:
      btmp: []

    # Matching on user name, using a default time range
    beacons:
      btmp:
        - users:
            gareth:
        - defaults:
            time_range:
                start: '8am'
                end: '4pm'

    # Matching on user name, overriding the default time range
    beacons:
      btmp:
        - users:
            gareth:
                time_range:
                    start: '8am'
                    end: '4pm'
        - defaults:
            time_range:
                start: '8am'
                end: '4pm'

    # Matching on group name, overriding the default time range
    beacons:
      btmp:
        - groups:
            users:
                time_range:
                    start: '8am'
                    end: '4pm'
        - defaults:
            time_range:
                start: '8am'
                end: '4pm'


Use Case: Posting Failed Login Events to Slack
==============================================

This can be done using the following reactor SLS:

.. code-block:: jinja

    report-wtmp:
      runner.salt.cmd:
        - args:
          - fun: slack.post_message
          - channel: mychannel      # Slack channel
          - from_name: someuser     # Slack user
          - message: "Failed login from `{{ data.get('user', '') or 'unknown user' }}` on `{{ data['id'] }}`"

Match the event like so in the master config file:

.. code-block:: yaml

    reactor:

      - 'salt/beacon/*/btmp/':
        - salt://reactor/btmp.sls

.. note::
    This approach uses the :py:mod:`slack execution module
    <salt.modules.slack_notify>` directly on the master, and therefore requires
    that the master has a slack API key in its configuration:

    .. code-block:: yaml

        slack:
          api_key: xoxb-XXXXXXXXXXXX-XXXXXXXXXXXX-XXXXXXXXXXXXXXXXXXXXXXXX

    See the :py:mod:`slack execution module <salt.modules.slack_notify>`
    documentation for more information. While you can use an individual user's
    API key to post to Slack, a bot user is likely better suited for this. The
    :py:mod:`slack engine <salt.engines.slack>` documentation has information
    on how to set up a bot user.
'''

# Import python libs
from __future__ import absolute_import, unicode_literals
import datetime
import logging
import os
import struct

# Import Salt Libs
import salt.utils.stringutils
import salt.utils.files

# Import 3rd-party libs
import salt.ext.six
# pylint: disable=import-error
from salt.ext.six.moves import map
# pylint: enable=import-error

__virtualname__ = 'btmp'
BTMP = '/var/log/btmp'
FMT = b'hi32s4s32s256shhiii4i20x'
FIELDS = [
          'type',
          'PID',
          'line',
          'inittab',
          'user',
          'hostname',
          'exit_status',
          'session',
          'time',
          'addr'
          ]
SIZE = struct.calcsize(FMT)
LOC_KEY = 'btmp.loc'

log = logging.getLogger(__name__)

# pylint: disable=import-error
try:
    import dateutil.parser as dateutil_parser
    _TIME_SUPPORTED = True
except ImportError:
    _TIME_SUPPORTED = False


def __virtual__():
    if os.path.isfile(BTMP):
        return __virtualname__
    return False


def _validate_time_range(trange, status, msg):
    '''
    Check time range
    '''
    # If trange is empty, just return the current status & msg
    if not trange:
        return status, msg

    if not isinstance(trange, dict):
        status = False
        msg = ('The time_range parameter for '
               'btmp beacon must '
               'be a dictionary.')

    if not all(k in trange for k in ('start', 'end')):
        status = False
        msg = ('The time_range parameter for '
               'btmp beacon must contain '
               'start & end options.')

    return status, msg


def _gather_group_members(group, groups, users):
    '''
    Gather group members
    '''
    _group = __salt__['group.info'](group)

    if not _group:
        log.warning('Group %s does not exist, ignoring.', group)
        return

    for member in _group['members']:
        if member not in users:
            users[member] = groups[group]


def _check_time_range(time_range, now):
    '''
    Check time range
    '''
    if _TIME_SUPPORTED:
        _start = dateutil_parser.parse(time_range['start'])
        _end = dateutil_parser.parse(time_range['end'])

        return bool(_start <= now <= _end)
    else:
        log.error('Dateutil is required.')
        return False


def _get_loc():
    '''
    return the active file location
    '''
    if LOC_KEY in __context__:
        return __context__[LOC_KEY]


def validate(config):
    '''
    Validate the beacon configuration
    '''
    vstatus = True
    vmsg = 'Valid beacon configuration'

    # Configuration for load beacon should be a list of dicts
    if not isinstance(config, list):
        vstatus = False
        vmsg = ('Configuration for btmp beacon must '
                'be a list.')
    else:
        _config = {}
        list(map(_config.update, config))

        if 'users' in _config:
            if not isinstance(_config['users'], dict):
                vstatus = False
                vmsg = ('User configuration for btmp beacon must '
                        'be a dictionary.')
            else:
                for user in _config['users']:
                    _time_range = _config['users'][user].get('time_range', {})
                    vstatus, vmsg = _validate_time_range(_time_range,
                                                         vstatus,
                                                         vmsg)

            if not vstatus:
                return vstatus, vmsg

        if 'groups' in _config:
            if not isinstance(_config['groups'], dict):
                vstatus = False
                vmsg = ('Group configuration for btmp beacon must '
                        'be a dictionary.')
            else:
                for group in _config['groups']:
                    _time_range = _config['groups'][group].get('time_range', {})
                    vstatus, vmsg = _validate_time_range(_time_range,
                                                         vstatus,
                                                         vmsg)
            if not vstatus:
                return vstatus, vmsg

        if 'defaults' in _config:
            if not isinstance(_config['defaults'], dict):
                vstatus = False
                vmsg = ('Defaults configuration for btmp beacon must '
                        'be a dictionary.')
            else:
                _time_range = _config['defaults'].get('time_range', {})
                vstatus, vmsg = _validate_time_range(_time_range,
                                                     vstatus,
                                                     vmsg)
            if not vstatus:
                return vstatus, vmsg

    return vstatus, vmsg


def beacon(config):
    '''
    Read the last btmp file and return information on the failed logins
    '''
    ret = []

    users = {}
    groups = {}
    defaults = None

    for config_item in config:
        if 'users' in config_item:
            users = config_item['users']

        if 'groups' in config_item:
            groups = config_item['groups']

        if 'defaults' in config_item:
            defaults = config_item['defaults']

    with salt.utils.files.fopen(BTMP, 'rb') as fp_:
        loc = __context__.get(LOC_KEY, 0)
        if loc == 0:
            fp_.seek(0, 2)
            __context__[LOC_KEY] = fp_.tell()
            return ret
        else:
            fp_.seek(loc)
        while True:
            now = datetime.datetime.now()
            raw = fp_.read(SIZE)
            if len(raw) != SIZE:
                return ret
            __context__[LOC_KEY] = fp_.tell()
            pack = struct.unpack(FMT, raw)
            event = {}
            for ind, field in enumerate(FIELDS):
                event[field] = pack[ind]
                if isinstance(event[field], salt.ext.six.string_types):
                    if isinstance(event[field], bytes):
                        event[field] = salt.utils.stringutils.to_unicode(event[field])
                    event[field] = event[field].strip('\x00')

            for group in groups:
                _gather_group_members(group, groups, users)

            if users:
                if event['user'] in users:
                    _user = users[event['user']]
                    if isinstance(_user, dict) and 'time_range' in _user:
                        if _check_time_range(_user['time_range'], now):
                            ret.append(event)
                    else:
                        if defaults and 'time_range' in defaults:
                            if _check_time_range(defaults['time_range'],
                                                 now):
                                ret.append(event)
                        else:
                            ret.append(event)
            else:
                if defaults and 'time_range' in defaults:
                    if _check_time_range(defaults['time_range'], now):
                        ret.append(event)
                else:
                    ret.append(event)
    return ret



# -*- coding: utf-8 -*-
'''
unit tests for the libvirt_events engine
'''
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import skipIf, TestCase
from tests.support.mock import (
    NO_MOCK,
    NO_MOCK_REASON,
    MagicMock,
    patch)

# Import Salt Libs
import salt.engines.libvirt_events as libvirt_events


# pylint: disable=protected-access,attribute-defined-outside-init,invalid-name,unused-argument,no-self-use


@skipIf(NO_MOCK, NO_MOCK_REASON)
class EngineLibvirtEventTestCase(TestCase, LoaderModuleMockMixin):
    '''
    Test cases for salt.engine.libvirt_events
    '''

    def setup_loader_modules(self):
        patcher = patch('salt.engines.libvirt_events.libvirt')
        self.mock_libvirt = patcher.start()
        self.mock_libvirt.getVersion.return_value = 2000000
        self.mock_libvirt.virEventRunDefaultImpl.return_value = -1  # Don't loop for ever
        self.mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE = 0
        self.mock_libvirt.VIR_DOMAIN_EVENT_ID_REBOOT = 1
        self.addCleanup(patcher.stop)
        self.addCleanup(delattr, self, 'mock_libvirt')
        return {libvirt_events: {}}

    @patch('salt.engines.libvirt_events.libvirt',
           VIR_PREFIX_NONE=0,
           VIR_PREFIX_ONE=1,
           VIR_PREFIX_TWO=2,
           VIR_PREFIX_SUB_FOO=0,
           VIR_PREFIX_SUB_BAR=1,
           VIR_PREFIX_SUB_FOOBAR=2)
    def test_get_libvirt_enum_string_subprefix(self, libvirt_mock):
        '''
        Make sure the libvirt enum value to string works reliably with
        elements with a sub prefix, eg VIR_PREFIX_SUB_* in this case.
        '''
        # Test case with a sub prefix

        assert libvirt_events._get_libvirt_enum_string('VIR_PREFIX_', 2) == 'two'

    @patch('salt.engines.libvirt_events.libvirt',
           VIR_PREFIX_FOO=0,
           VIR_PREFIX_BAR_FOO=1)
    def test_get_libvirt_enum_string_underscores(self, libvirt_mock):
        '''
        Make sure the libvirt enum value to string works reliably and items
        with an underscore aren't confused with sub prefixes.
        '''
        assert libvirt_events._get_libvirt_enum_string('VIR_PREFIX_', 1) == 'bar foo'

    @patch('salt.engines.libvirt_events.libvirt',
           VIR_DOMAIN_EVENT_CRASHED_PANICKED=0,
           VIR_DOMAIN_EVENT_DEFINED=0,
           VIR_DOMAIN_EVENT_UNDEFINED=1,
           VIR_DOMAIN_EVENT_CRASHED=2,
           VIR_DOMAIN_EVENT_DEFINED_ADDED=0,
           VIR_DOMAIN_EVENT_DEFINED_UPDATED=1)
    def test_get_domain_event_detail(self, mock_libvirt):
        '''
        Test get_domain_event_detail function
        '''
        assert libvirt_events._get_domain_event_detail(1, 2) == ('undefined', 'unknown')
        assert libvirt_events._get_domain_event_detail(0, 1) == ('defined', 'updated')
        assert libvirt_events._get_domain_event_detail(4, 2) == ('unknown', 'unknown')

    @patch('salt.engines.libvirt_events.libvirt', VIR_NETWORK_EVENT_ID_LIFECYCLE=1000)
    def test_event_register(self, mock_libvirt):
        '''
        Test that the libvirt_events engine actually registers events catch them and cleans
        before leaving the place.
        '''
        mock_cnx = MagicMock()
        mock_libvirt.openReadOnly.return_value = mock_cnx

        # Don't loop for ever
        mock_libvirt.virEventRunDefaultImpl.return_value = -1

        mock_cnx.networkEventRegisterAny.return_value = 10000

        libvirt_events.start('test:///', 'test/prefix')

        # Check that the connection has been opened
        mock_libvirt.openReadOnly.assert_called_once_with('test:///')

        # Check that the connection has been closed
        mock_cnx.close.assert_called_once()

        # Check events registration and deregistration
        mock_cnx.domainEventRegisterAny.assert_any_call(
                None, mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE,
                libvirt_events._domain_event_lifecycle_cb,
                {'prefix': 'test/prefix', 'object': 'domain', 'event': 'lifecycle'})
        mock_cnx.networkEventRegisterAny.assert_any_call(
                None, mock_libvirt.VIR_NETWORK_EVENT_ID_LIFECYCLE,
                libvirt_events._network_event_lifecycle_cb,
                {'prefix': 'test/prefix', 'object': 'network', 'event': 'lifecycle'})

        # Check that the deregister events are called with the result of register
        mock_cnx.networkEventDeregisterAny.assert_called_with(
                mock_cnx.networkEventRegisterAny.return_value)

        # Check that the default 'all' filter actually worked
        counts = {obj: len(callback_def) for obj, callback_def in libvirt_events.CALLBACK_DEFS.items()}
        for obj, count in counts.items():
            register = libvirt_events.REGISTER_FUNCTIONS[obj]
            assert getattr(mock_cnx, register).call_count == count

    def test_event_skipped(self):
        '''
        Test that events are skipped if their ID isn't defined in the libvirt
        module (older libvirt)
        '''
        self.mock_libvirt.mock_add_spec([
                'openReadOnly',
                'virEventRegisterDefaultImpl',
                'virEventRunDefaultImpl',
                'VIR_DOMAIN_EVENT_ID_LIFECYCLE'], spec_set=True)

        libvirt_events.start('test:///', 'test/prefix')

        # Check events registration and deregistration
        mock_cnx = self.mock_libvirt.openReadOnly.return_value

        mock_cnx.domainEventRegisterAny.assert_any_call(
                None, self.mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE,
                libvirt_events._domain_event_lifecycle_cb,
                {'prefix': 'test/prefix', 'object': 'domain', 'event': 'lifecycle'})

        # Network events should have been skipped
        mock_cnx.networkEventRegisterAny.assert_not_called()

    def test_event_filtered(self):
        '''
        Test that events are skipped if their ID isn't defined in the libvirt
        module (older libvirt)
        '''
        libvirt_events.start('test', 'test/prefix', 'domain/lifecycle')

        # Check events registration and deregistration
        mock_cnx = self.mock_libvirt.openReadOnly.return_value

        mock_cnx.domainEventRegisterAny.assert_any_call(
                None, 0, libvirt_events._domain_event_lifecycle_cb,
                {'prefix': 'test/prefix', 'object': 'domain', 'event': 'lifecycle'})

        # Network events should have been filtered out
        mock_cnx.networkEventRegisterAny.assert_not_called()



# -*- coding: utf-8 -*-
'''
Tests for the nacl execution module
'''
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
import salt.utils.stringutils
import salt.modules.nacl as nacl

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase
from tests.support.unit import skipIf

try:
    import libnacl.secret  # pylint: disable=unused-import
    import libnacl.sealed  # pylint: disable=unused-import
    HAS_LIBNACL = True
except ImportError:
    HAS_LIBNACL = False


@skipIf(not HAS_LIBNACL, 'skipping test_nacl, libnacl is unavailable')
class NaclTest(TestCase, LoaderModuleMockMixin):
    '''
    Test the nacl runner
    '''
    def setup_loader_modules(self):
        self.unencrypted_data = salt.utils.stringutils.to_bytes('hello')
        self.opts = salt.config.DEFAULT_MINION_OPTS
        utils = salt.loader.utils(self.opts)
        funcs = salt.loader.minion_mods(self.opts, utils=utils, whitelist=['nacl'])

        return {
            nacl: {
                '__opts__': self.opts,
                '__utils__': utils,
                '__salt__': funcs
            },
        }

    def setUp(self):
        # Generate the keys
        ret = nacl.keygen()
        self.assertIn('pk', ret)
        self.assertIn('sk', ret)
        self.pk = ret['pk']
        self.sk = ret['sk']

    def test_keygen(self):
        '''
        Test keygen
        '''
        self.assertEqual(len(self.pk), 44)
        self.assertEqual(len(self.sk), 44)

    def test_enc_dec(self):
        '''
        Generate keys, encrypt, then decrypt.
        '''
        # Encrypt with pk
        encrypted_data = nacl.enc(data=self.unencrypted_data, pk=self.pk)

        # Decrypt with sk
        decrypted_data = nacl.dec(data=encrypted_data, sk=self.sk)
        self.assertEqual(self.unencrypted_data, decrypted_data)

    def test_sealedbox_enc_dec(self):
        '''
        Generate keys, encrypt, then decrypt.
        '''
        # Encrypt with pk
        encrypted_data = nacl.sealedbox_encrypt(
            data=self.unencrypted_data, pk=self.pk)

        # Decrypt with sk
        decrypted_data = nacl.sealedbox_decrypt(
            data=encrypted_data, sk=self.sk)

        self.assertEqual(self.unencrypted_data, decrypted_data)

    def test_secretbox_enc_dec(self):
        '''
        Generate keys, encrypt, then decrypt.
        '''
        # Encrypt with sk
        encrypted_data = nacl.secretbox_encrypt(
            data=self.unencrypted_data, sk=self.sk)

        # Decrypt with sk
        decrypted_data = nacl.secretbox_decrypt(
            data=encrypted_data, sk=self.sk)

        self.assertEqual(self.unencrypted_data, decrypted_data)
